/*
    Copyright (C) 2020 Jiri Pittner <jiri@pittnerovi.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "ahrs2.h"

using namespace LA_Quaternion;
using namespace LA_Vecmat3;



/*new IMU orientation algorithm suitable for avionics
based on:
"Improved Formulation of the IMU and MARG Orientation Gradient Descent Algorithm for Motion Tracking in Human-Machine Interfaces" by Marcel Admiraal, Samuel Wilson, Ravi Vaidyanathan, in 2017 IEEE International Conference on Multisensor Fusion and Integration for Intelligent Systems (MFI 2017)
"Formulation of a new gradient descent MARG orientation algorithm: Case study on robot teleoperation" by
Samuel Wilson, Henry Eberle, Yoshikatsu Hayashi, Sebastian O.H. Madgwick, Alison McGregor, Xingjian Jing, Ravi Vaidyanathan in Mechanical Systems and Signal Processing 130, 183-200 (2019).
The treatment of non-inertial forces for in-flight operation has been implemented based on the description in the standard textbook on avionics: R.P.G. Collinson, "Introduction to Avionics Systems", 3rd edition, Springer, Dordrecht Heidelberg London New York 2011 ISBN 978-94-007-0707-8 DOI 10.1007/978-94-007-0708-5.
*/

//
//make sure that all inputs are in SI units
//this algorithm is general (not optimized for sparse vectors of the references)
//inertial velocity and acceleration are with respect to stationary frame with axes oriented as the body frame at a given time instant, not with respect to earth frame orientation!
//omega is locally measured with respect to instantaneous body frame
//
template<typename T>
void AHRSupdate(Quaternion<T> &orientation,const T dt,const T beta, const T beta2, const Vec3<T> &inertial_velocity, const Vec3<T> &linear_acceleration, const Vec3<T> &omega, const Vec3<T> &acceleration, const Vec3<T> &magnet , const Vec3<T> &gravity_reference,  const Vec3<T> &magnetic_reference, bool normalize_gradcorr1, bool normalize_gradcorr2, bool ref2_crossproduct,const Vec3<T> *ref1_override)
{
//correct measured accelerations for non-inertial forces in the linearly-accelerating and rotating body-frame
//normalize the reference and measured vectors

Vec3<T> ref1_body;
if(ref1_override) ref1_body= *ref1_override; //use externally lowpassed value
else ref1_body = acceleration - linear_acceleration - omega * inertial_velocity; 
ref1_body.normalize();
Vec3<T> ref1_earth = gravity_reference;
ref1_earth.normalize();

//build effective second reference and measured vector as a cross product of gravity with magnetic field
Vec3<T> ref2_body = ref2_crossproduct? ref1_body * magnet: magnet;
ref2_body.normalize();
Vec3<T> ref2_earth = ref2_crossproduct? gravity_reference * magnetic_reference: magnetic_reference;
ref2_earth.normalize();

//compute time derivative of the orientation quaternion from angular velocity
Quaternion<T> qder;
qder = orientation.times_vec3(omega); //this step matches original Madgwick's code
qder *= 0.5f;

//correct by the gradient descent algorithm for the two references
Vec3<T> ref1_rot, ref2_rot;
Quaternion<T> qcorr = orientation.rotate_match(ref1_rot,ref1_earth,ref1_body);
Quaternion<T> qcorr2 = orientation.rotate_match(ref2_rot,ref2_earth,ref2_body);
if(normalize_gradcorr1) qcorr.normalize(); 
if(normalize_gradcorr2) qcorr2.normalize(); 
qder -= qcorr * beta;
qder -= qcorr2 * beta2;

//time-propagate and renormalize
orientation += qder * dt;
orientation.normalize();
}




#define INSTANTIZE(T) \
template void AHRSupdate(Quaternion<T> &orientation,const T dt,const T beta, const T beta2, const Vec3<T> &inertial_velocity, const Vec3<T> &linear_acceleration, const Vec3<T> &omega, const Vec3<T> &acceleration, const Vec3<T> &magnet, const Vec3<T> &gravity_reference,  const Vec3<T> &magnetic_reference, bool normalize_gradcorr1, bool normalize_gradcorr2, bool ref2_crossproduct, const Vec3<T> *ref1_override); \



INSTANTIZE(float)
#ifndef QUAT_NO_DOUBLE
INSTANTIZE(double)
#endif


