/*
    Copyright (C) 2020 Jiri Pittner <jiri@pittnerovi.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include <stdio.h>
#include <math.h>
#include "rotcoder.h"
#include "vecmat3.h"
#include "quaternion.h"
#include "avi_gui.h"
#include "avi_calibration.h"
#include "magnetometer.h"
#include "avionics.h"

T6963_STM32 *mylcd;
WIDGET *activewidget;

bool displaymode_changed=true;

//for user-setup of instrument orientation
//@@@@ with the final gui, these will not be stored here,
//but a pointer to them will be stored in a widget class

volatile DISPLAY_MODE display_mode=gui_mode;
static WIDGET **gui;

//C-linkage rotcoder callbacks
extern "C" {

void process_press(uint16_t ms)
{
if(display_mode==gui_mode && activewidget) activewidget->process_press(ms);
}

void process_rotation(uint8_t dir)
{
if(display_mode!=gui_mode) return;
if(activewidget) activewidget->process_rotation(dir);
}

void process_button(uint8_t which, uint16_t ms)
{
printf("button %d pressed for %dms\n",which,ms);
switch(which) {
	case 0: //left
		display_mode= (DISPLAY_MODE) ((display_mode+1)%N_DISPLAY_MODE);
		switch(display_mode) {
                        case avitext: //debug
                                break;
                        case gui_mode: //gui
                                break;
                        case magnet_cal: //compass calibration
                                break;
                        case gyro_cal: //gyro calibration
                                break;
			case can_debug: //can
				break;
		}
		break;
	case 1: //right
		switch(display_mode) {
			case avitext: //debug
				break;
			case gui_mode: //gui
				break;
			case magnet_cal: //compass calibration
				fit_magnetic_calibration(HAS_EEPROM,1); //and preserve data for continuous recalibration
				{
				char z[64];
				sprintf(z,"new magoff %f %f %f\n",magnetic_offsets[0],magnetic_offsets[1],magnetic_offsets[2]);
				if(logfile) fputs(z,logfile);
				sprintf(z,"new magsc %f %f %f",magnetic_scales[0],magnetic_scales[1],magnetic_scales[2]);
				if(logfile) fputs(z,logfile);
				}
				break;
			case gyro_cal: //gyro and pitot calibration
                                store_omega_calibration(0);
				{
                                char z[64];
				sprintf(z,"new pitot offsets %f %f\n",p_pitot_calibration_offset,p_pitot2_calibration_offset);
                                if(logfile) fputs(z,logfile);
                                mylcd->setCursor(0,14);mylcd->write(z);
                                sprintf(z,"New offs %f %f %f\n",omega_offset_x,omega_offset_y,omega_offset_z);
				if(logfile) fputs(z,logfile);
                                mylcd->setCursor(0,15);mylcd->write(z);
                                }
				break;
			case can_debug:
				break;
		}
		break;
	case 2: //middle
		switch(display_mode) {
                        case avitext: //debug
                                break;
                        case gui_mode: //gui
                                break;  
                        case magnet_cal: //compass calibration
				reset_magnetic_calibration(HAS_EEPROM);
				init_magnetic_calibration(); //erase old calibration data
				mylcd->clearGraphic();
				displaymode_changed=true; //clear displayed points
				mylcd->setCursor(0,2);mylcd->write("Calibration reset to default");
                                break;
                        case gyro_cal: //gyro calibration
				store_omega_calibration(1);
				init_omega_calibration();
                                mylcd->setCursor(0,2);mylcd->write("Calibration reset to default");
                                break;
			case can_debug:
				break;
                }
		break;
	}
}

}//extern "C" 

void textwidget::draw(bool firsttime) //here firsttime has no effect
             {
             if(!text) return; 
             char z[64];
             if(number) sprintf(z,text,*number * scale);
             mylcd->multiwrite(x/TEXTFONT_X,y/TEXTFONT_Y,(number?z:text));
             }

textwidget::textwidget(uint16_t x0,uint16_t y0, uint8_t sizex0, uint8_t sizey0, const char *text0, volatile float *number0, const float scale0)
		: widget(x0,y0,sizex0,sizey0),
		scale(scale0), text(text0),number(number0)
                {}

tunable_textwidget::tunable_textwidget(uint16_t x0,uint16_t y0,  uint8_t sizex0, uint8_t sizey0,  const char *text0, volatile float *number0, const float scale0, const float step0, const float defaultv0, const float mini0, const float maxi0, volatile bool *changed0, const bool multiplicative0)
	: textwidget(x0,y0,sizex0,sizey0,text0,number0,scale0), 
	step(step0), defaultv(defaultv0), mini(mini0), maxi(maxi0), changed(changed0), multiplicative(multiplicative0)
	{ }

void tunable_textwidget::process_press(uint16_t ms)
{
if(changed) *changed=true;
*number = defaultv;
}

void tunable_textwidget:: process_rotation(uint8_t dir)
{
if(changed) *changed=true;
if(multiplicative)
	{
        if(dir) *number *= step; else *number /= step;
        }

else
	{
	if(dir) *number += step; else *number -= step;
	}
if(*number < mini) *number=mini;
if(*number > maxi) *number=maxi;
}

void tunable_textwidget::activate(bool on)
{
mylcd->drawRect(x,y,sizex,sizey,on);
}


ballwidget::ballwidget(uint16_t x0,uint16_t y0,  uint8_t sizex0, uint8_t sizey0, float *y_acceleration0, float sensitivity0)
                : widget(x0,y0,sizex0,sizey0), 
		  map(sizex0,sizey0),
		  y_acceleration(y_acceleration0), sensitivity(sensitivity0)
                {}

void ballwidget::draw(bool firsttime)
{
float tmp= - *y_acceleration/sensitivity;
if(tmp>1.f) tmp=1.;
if(tmp< -1.f) tmp= -1.;
uint16_t xorigin= x+sizex/2+ .5*tmp*(sizex-sizey);
uint16_t yorigin= y+sizey/2;
uint16_t radius=sizey/2-1;
uint16_t xstart = xorigin-radius;
uint16_t xend= xorigin+radius;
uint16_t radius2 = radius*radius;


#if 0
//old version
//draw simple rectangle to avoid a need for double buffering to prevent flicker
if(xstart>x+1) mylcd->drawFillRect(x+1,y+1,xstart-1,y+sizey-2,0); //erase left
mylcd->drawFillRect(xstart,y+1,xend,y+sizey-2,1); //draw square "ball"
if(xend<x+sizex) mylcd->drawFillRect(xend+1,y+1,x+sizex-1,y+sizey-2,0); //erase right

#else
//loop over interior pixels of the rectangle and decide which are inside the ball and which are outside
for(uint16_t yy=1; yy<sizey-1; ++yy)
	{
	uint16_t ytrue=y+yy;
	int16_t yoffset2 = ytrue-yorigin; yoffset2 *= yoffset2;
	for(uint16_t xx=1; xx<sizex-1; ++xx)
		{
		uint16_t xtrue=x+xx;
		int16_t xoffset2 = xtrue-xorigin; xoffset2 *= xoffset2;
		bool inside = yoffset2+xoffset2 <= radius2;
		if(firsttime || inside!=map(xx,yy))
			{
			map.assign(xx,yy,inside);
			mylcd->drawPixel(xtrue,ytrue,inside);
			}
		}
	}
#endif


if(firsttime) mylcd->drawRect(x,y,sizex,sizey,1); //frame around
}



horizonwidget::horizonwidget(uint16_t x0,uint16_t y0,  uint8_t sizex0, uint8_t sizey0, float *pitch0, float *roll0, float pitchscale0)
	: widget(x0,y0,sizex0,sizey0), 
	  map(sizex0,sizey0),
	  pitch(pitch0), roll(roll0), pitchscale(pitchscale0)
	{}


void horizonwidget::draw(bool firsttime)
{
float degs = *roll *M_PI/180.f;
float sinus = -sinf(degs); //negative due to display y growing down
float cosinus = cosf(degs); 

//center of the widget
uint16_t x00= x+sizex/2;
uint16_t y00=y+sizey/2;

//determine center of the horizon line
int32_t x0 = x00 - *pitch*pitchscale*sinus;
int32_t y0 = y00 + *pitch*pitchscale*cosinus; 

//make a nice horizontal line for negligible roll
if(fabsf(sinus)<1.f/sizex) sinus=0; 
if(fabsf(cosinus)<1.f/sizey) cosinus=0; 

uint16_t ix,iy;
//loop over all pixels and decide whether they are below horizon or not
for(iy=y+1; iy<y+sizey-1; ++iy)
	{
	for(ix=x+1; ix<x+sizex-1; ++ix)
		{
		//determine if pixel is under the horizon
		bool below;
		if(*pitch>45.f) below=0;
		else if(*pitch<-45.f) below=1;
		else if(cosinus*(iy-y0)  <  sinus*(ix-x0)) below=0;
		else below=1;

		//invert crosshair bits
		if(iy==y00)
			{
			if(
				ix>= x00-8 && ix <= x00-2 ||
				ix>= x00+2 && ix <= x00+8 
			  )
				below ^= 1;
			}

		if(firsttime || below!=map(ix-x,iy-y))
			{
			map.assign(ix-x,iy-y,below);
			mylcd->drawPixel(ix,iy,below);
			}
		}
	}


if(firsttime) mylcd->drawRect(x,y,sizex,sizey,1); //frame it
}



void init_gui(T6963_STM32 *lcd, WIDGET **gui0)
{
mylcd=lcd;
activewidget=NULL;
mylcd->fastMode(true); //hardware does not support reading from LCD anyway
mylcd->setDispMode(true,true,false,false); //activate both text and graphics regime
mylcd->setBacklight(true);
mylcd->begin(LCD_X,LCD_Y); //default 6x8 chars gives 16x40 text size 
mylcd->clearText();
mylcd->clearGraphic();
printf("LCD init done\n");

if(gui0==NULL) return;

//init widget list
gui=gui0;

//sanity check for widget intersections
WIDGET **p=gui;
while(*p)
        {
	uint16_t x1= (*p)->x;
	uint16_t y1= (*p)->y;
	uint16_t x2= x1-1+(*p)->sizex;
	uint16_t y2= y1-1+(*p)->sizey;
	WIDGET **q = p+1;
	while(*q)
		{
		uint16_t qx1= (*q)->x;
	        uint16_t qy1= (*q)->y;
       	 	uint16_t qx2= qx1-1+(*q)->sizex;
        	uint16_t qy2= qy1-1+(*q)->sizey;
		if(
			qx1 >= x1 && qx1 <= x2 && qy1 >= y1 && qy1 <= y2  ||
			qx1 >= x1 && qx1 <= x2 && qy2 >= y1 && qy2 <= y2  ||
			qx2 >= x1 && qx2 <= x2 && qy1 >= y1 && qy1 <= y2  ||
			qx2 >= x1 && qx2 <= x2 && qy2 >= y1 && qy2 <= y2 
		  ) printf("ERROR: intersecting widgets %d %d\n",p-gui, q-gui);
		++q;
		}
	++p;
	}

init_rotcoder_buttons(process_press,process_rotation,process_button);
}


void process_touch(float x, float y)
{
//printf("TOUCH @ %f %f\n",x,y);
if(display_mode!=gui_mode) return;

uint16_t x0 = x*LCD_X;
uint16_t y0 = y*LCD_Y;

//go through the widget list and find closest to the touch point
activewidget=NULL;
WIDGET **p=gui;
while(*p) 
	{
	uint16_t x1= (*p)->x;
	uint16_t x2= x1+(*p)->sizex;
	uint16_t y1= (*p)->y;
        uint16_t y2= y1+(*p)->sizey;
	if(x0>=x1 && x0<=x2 && y0>=y1 && y0<=y2)
		{
		//printf("widget found\n");
		activewidget = *p;
		break;
		}
	++p;
	}
}


void update_gui(T6963_STM32 *lcd, bool firsttime)
{
//go through the widget list and draw all
WIDGET **p=gui;
while(*p)
	{
	//frame box if activated by touch
	if(*p != activewidget) (*p)->activate(0);
	(*p)->draw(firsttime);
	if(*p == activewidget) (*p)->activate(1);
	++p;
	}
}

void xyplot(float x, float y, float scale)
{
uint8_t ix,iy;
ix=120+x*scale;
iy=64+y*scale;
mylcd->drawPixel(ix,iy,1);
}

