/*
    Copyright (C) 2022 Jiri Pittner <jiri@pittnerovi.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

//hardware: kicad caninterface.pro
//AF9 CAN_RX = PB8 CAN_TX = PB9 CAN_EN = PB5(low active)

#include "can.h"

uint8_t init_can(uint32_t baud,uint8_t autoretransmission)
{
//Enable the CAN controller interface clock using
RCC_APB1PeriphClockCmd(RCC_APB1Periph_CAN1, ENABLE);

//configure pins
RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOB, ENABLE);
GPIO_PinAFConfig(GPIOB, GPIO_PinSource8, GPIO_AF_9);
GPIO_PinAFConfig(GPIOB, GPIO_PinSource9, GPIO_AF_9);

GPIO_InitTypeDef GPIO_InitStructure;
GPIO_InitStructure.GPIO_Pin =  GPIO_Pin_5;
GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
GPIO_Init(GPIOB, &GPIO_InitStructure);

GPIO_InitStructure.GPIO_Pin =  GPIO_Pin_8|GPIO_Pin_9;
GPIO_InitStructure.GPIO_Mode = GPIO_Mode_AF;
GPIO_Init(GPIOB, &GPIO_InitStructure);

//initialize CAN
CAN_InitTypeDef CAN_InitStruct;
CAN_StructInit(&CAN_InitStruct);//defaults

//set timing parameters
CAN_InitStruct.CAN_SJW = CAN_SJW_1tq;
CAN_InitStruct.CAN_BS1 = CAN_BS1_2tq;
CAN_InitStruct.CAN_BS2 = CAN_BS2_1tq;

/* Initialize the no automatic retransmission */
if(autoretransmission)  CAN_InitStruct.CAN_NART = DISABLE;
else  CAN_InitStruct.CAN_NART = ENABLE;

uint32_t denom=(1+2+1)*2*baud;
if((XTAL*PLLMULT)%denom) return 2; //timing must be exact
uint16_t ps = (XTAL*PLLMULT)/denom;
if(ps>1023) return 2;
if(ps==0) return 2;
CAN_InitStruct.CAN_Prescaler = ps; 

if(CAN_Init(CAN1,&CAN_InitStruct)!=CAN_InitStatus_Success) return 1;

//initialize reception filter - mandatory
CAN_FilterInitTypeDef CAN_FilterInitStruct;
CAN_FilterInitStruct.CAN_FilterIdHigh=0;
CAN_FilterInitStruct.CAN_FilterIdLow=0;
CAN_FilterInitStruct.CAN_FilterMaskIdHigh=0;
CAN_FilterInitStruct.CAN_FilterMaskIdLow=0;
CAN_FilterInitStruct.CAN_FilterFIFOAssignment = CAN_Filter_FIFO0;
CAN_FilterInitStruct.CAN_FilterNumber=0;
CAN_FilterInitStruct.CAN_FilterMode=CAN_FilterMode_IdMask;
CAN_FilterInitStruct.CAN_FilterScale=CAN_FilterScale_32bit;
CAN_FilterInitStruct.CAN_FilterActivation=ENABLE;
CAN_FilterInit(&CAN_FilterInitStruct);
CAN_FilterInitStruct.CAN_FilterFIFOAssignment = CAN_Filter_FIFO1;
CAN_FilterInit(&CAN_FilterInitStruct);


cbi(GPIOB,5); //enable CAN transceiver
return 0;
}

void can_rx_filter(uint32_t id, uint32_t id_mask)
{
id<<=21;
id_mask<<=21;

CAN_FilterInitTypeDef CAN_FilterInitStruct;
CAN_FilterInitStruct.CAN_FilterIdHigh=id >>16;
CAN_FilterInitStruct.CAN_FilterIdLow=id&0xffff;
CAN_FilterInitStruct.CAN_FilterMaskIdHigh=id_mask>>16;
CAN_FilterInitStruct.CAN_FilterMaskIdLow=id_mask&0xffff;
CAN_FilterInitStruct.CAN_FilterFIFOAssignment = CAN_Filter_FIFO0;
CAN_FilterInitStruct.CAN_FilterNumber=0;
CAN_FilterInitStruct.CAN_FilterMode=CAN_FilterMode_IdMask;
CAN_FilterInitStruct.CAN_FilterScale=CAN_FilterScale_32bit;
CAN_FilterInitStruct.CAN_FilterActivation=ENABLE;
CAN_FilterInit(&CAN_FilterInitStruct);
CAN_FilterInitStruct.CAN_FilterFIFOAssignment = CAN_Filter_FIFO1; 
CAN_FilterInit(&CAN_FilterInitStruct);
}



//routines using the STD peripheral library, unnecessarily copy data

uint8_t can_receive2(uint8_t *idtype,uint32_t *stdid, uint32_t *extid, uint8_t *rtr,uint8_t *data)
{
//try both fifos
uint8_t fifo=CAN_FIFO0;
if(CAN_MessagePending(CAN1,fifo)==0)
	{
	fifo=CAN_FIFO1;
	if(CAN_MessagePending(CAN1,fifo)==0) return 0;
	}
CanRxMsg RxMessage;
CAN_Receive(CAN1,fifo, &RxMessage);
*rtr=RxMessage.RTR;
*idtype=RxMessage.IDE;
*stdid=RxMessage.StdId;
if(RxMessage.IDE==CAN_Id_Extended)  *extid=RxMessage.ExtId;
memcpy(data,RxMessage.Data,8);
CAN_FIFORelease(CAN1,fifo);
return RxMessage.DLC;
}


void can_transmit2_ext(uint16_t id, uint32_t extid, uint8_t rtr, uint8_t len, uint8_t *data)
{
CanTxMsg TxMessage;
TxMessage.StdId = id; //0-7ff
TxMessage.ExtId =extid;
TxMessage.IDE =CAN_Id_Extended;
if(rtr) TxMessage.RTR = CAN_RTR_Remote;
else TxMessage.RTR = CAN_RTR_Data;
TxMessage.DLC=len;
memcpy(TxMessage.Data,data,8);
CAN_Transmit(CAN1,&TxMessage);
}

void can_transmit2(uint16_t id, uint8_t rtr, uint8_t len, uint8_t *data)
{
CanTxMsg TxMessage;
TxMessage.StdId = id; //0-7ff
TxMessage.IDE =CAN_Id_Standard;
if(rtr) TxMessage.RTR = CAN_RTR_Remote;
else TxMessage.RTR = CAN_RTR_Data;
TxMessage.DLC=len;
memcpy(TxMessage.Data,data,8);
CAN_Transmit(CAN1,&TxMessage);
}

//faster versions without memcpy and using my can structure


//receive -only standard 11bit IDs
uint8_t can_receive(CAN_MESSAGE *m)
{
//try both fifos
uint8_t fifo=CAN_FIFO0;
if(CAN_MessagePending(CAN1,fifo)==0)
        {
        fifo=CAN_FIFO1;
        if(CAN_MessagePending(CAN1,fifo)==0) return 0;
        }

  uint8_t IDE = (uint8_t)0x04 & CAN1->sFIFOMailBox[fifo].RIR;
  if(IDE != CAN_Id_Standard) { m->length=0; goto done;}
  m->id = (uint32_t)0x000007FF & (CAN1->sFIFOMailBox[fifo].RIR >> 21);
  m->length = (uint8_t)0x0F & CAN1->sFIFOMailBox[fifo].RDTR;
  m->data.reg[0] = CAN1->sFIFOMailBox[fifo].RDLR;
  m->data.reg[1] = CAN1->sFIFOMailBox[fifo].RDHR;

done: /* Release the FIFO */
  if (fifo == CAN_FIFO0) CAN1->RF0R |= CAN_RF0R_RFOM0;
  else CAN1->RF1R |= CAN_RF1R_RFOM1;

return m->length;
}


#define TMIDxR_TXRQ       ((uint32_t)0x00000001) /* Transmit mailbox request */
uint8_t can_transmit(CAN_MESSAGE *m,uint8_t rtr)
{
uint8_t transmit_mailbox = 0;
  /* Select one empty transmit mailbox */
  if ((CAN1->TSR&CAN_TSR_TME0) == CAN_TSR_TME0)
  {
    transmit_mailbox = 0;
  } 
  else if ((CAN1->TSR&CAN_TSR_TME1) == CAN_TSR_TME1)
  {
    transmit_mailbox = 1;
  }
  else if ((CAN1->TSR&CAN_TSR_TME2) == CAN_TSR_TME2)
  {
    transmit_mailbox = 2;
  }
  else
    return -1;

    /* Set up the Id */
    CAN1->sTxMailBox[transmit_mailbox].TIR &= TMIDxR_TXRQ;
    CAN1->sTxMailBox[transmit_mailbox].TIR |= ((m->id << 21) | (rtr?CAN_RTR_Remote:CAN_RTR_Data));

    /* Set up the DLC */
    CAN1->sTxMailBox[transmit_mailbox].TDTR &= (uint32_t)0xFFFFFFF0;
    CAN1->sTxMailBox[transmit_mailbox].TDTR |= (m->length &0xf);
    CAN1->sTxMailBox[transmit_mailbox].TDLR = m->data.reg[0];
    CAN1->sTxMailBox[transmit_mailbox].TDHR = m->data.reg[1];
    /* Request transmission */
    CAN1->sTxMailBox[transmit_mailbox].TIR |= TMIDxR_TXRQ;
return 0;
}
