/*
    Copyright (C) 2022 Jiri Pittner <jiri@pittnerovi.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdint.h>

#include "can_devind.h"
#include "obd2.h"

#define XSTR(s) STR(s)
#define STR(s) #s

#ifdef WITH_DBC
#include XSTR(DBC_INCLUDE)
#endif

#define DBC_FUNCTION_(X) decode_can_message_##X
#define DBC_FUNCTION(X) DBC_FUNCTION_(X) 



//see https://en.wikipedia.org/wiki/OBD-II_PIDs

static inline char bin2letter(uint8_t b)
{
return  b<10? '0'+b : 'A'+b-10;
}



//DTC need special treatment - possibly multiple responses, up to 3 2-byte codes in a msg
char *dtc2text(uint8_t *code, char *dtc)
{
const char codes[5]="PCBU";
dtc[0]= codes[code[0]>>6];
dtc[1]= '0' + ((code[0]>>4)&3);
dtc[2]= bin2letter(code[0]&0x0f);
dtc[3]=  bin2letter(code[1]>>4);
dtc[3]= bin2letter(code[1]&0x0f);
dtc[6]= 0;
return dtc;
}



//returns  number of filled elements in CAN_VALUE array, 0 or negative for error
int8_t decode_can_message(CAN_MESSAGE *m, CAN_VALUE *v)
{
int8_t r= -20;
#ifdef WITH_OBD2
if(m->id >= 0x7e8 && m->id <=0x7ef)  //OBD2 reply, SAE standartized
	{
	if(m->length>8) return -30; //extended message; use a special routine to read VIN
	switch(m->data.byte[1])  //service
		{
		case 0x41:// S1
		case 0x42:// S2 //implemented via dbc file
			m->id =  0x7e8; //to be recognized in the generated routine
			r=decode_can_message_obd2(m,m->data.byte[2],v);
			break;
		case 0x43:// S3
		case 0x47:// S7
		case 0x4a:// SA
			{
			v[0].value.uint8[0]=m->data.byte[3];
			v[0].value.uint8[1]=m->data.byte[4];
			v[0].type=3;
                        v[1].value.uint8[0]=m->data.byte[5];
                        v[1].value.uint8[1]=m->data.byte[6];
                        v[1].type=3;
                        v[2].value.uint8[0]=m->data.byte[7];
                        v[2].value.uint8[1]=m->data.byte[8];
                        v[2].type=3;
			r=3;
			}
			break;
		case 0x44: //S4;
			r= 0;
			break;
		case 0x49: //S9 
			r= -5; //unsupported
			break;
		default: 
			return -10;
		}
	}
else //unsolicited message, manufacturer&model-specific
	{
#endif
#ifdef WITH_DBC
	r=DBC_FUNCTION(WITH_DBC)(m,0,v);
#endif
#ifdef WITH_OBD2
	}
#endif
return r;
}


void sprintf_can_value(CAN_VALUE *v,  char *line)
{
switch(v->type)
	{
	case 3: //DTC code
		{
		char tmp[8];
		sprintf(line,"%5s",dtc2text(v->value.uint8,tmp));
		}
		break;
	case 2: //float
		sprintf(line,"%f %s",v->value.floatval,v->unit);
                break;
	case 1: //unsigned int
		sprintf(line,
#ifdef ARM_SOURCE37
			"%lu %s"
#else
			"%u %s"
#endif
			,v->value.uint32,v->unit);
                break;
	case -1: //signed int
		sprintf(line,
#ifdef ARM_SOURCE37
			"%ld %s"
#else
			"%d %s"
#endif
			,v->value.int32,v->unit);
                break;
	case 0: //raw
	default: 
		sprintf(line,
#ifdef ARM_SOURCE37
			"%08lx order ABCD"
#else
			"%08x order ABCD"
#endif
			,v->value.uint32);
		break;
	}
}

void sprintf_can_name_value(CAN_VALUE *v,  char *line)
{
sprintf(line,"%s %s = ",v->groupname,v->signame);
sprintf_can_value(v,line+strlen(line));
}


void decode_and_output_can_message(CAN_MESSAGE *m, FILE *out)
{
CAN_VALUE val[MAXSIG];
char groupnames[MAXSIG*SIGNAME_LEN];
char signames[MAXSIG*SIGNAME_LEN];
char units[MAXSIG*UNIT_NAME_LEN];
uint8_t  i;
for(i=0; i<MAXSIG; ++i)
	{
	val[i].groupname=groupnames+SIGNAME_LEN*i;
	val[i].signame=signames+SIGNAME_LEN*i;
	val[i].unit=units+UNIT_NAME_LEN*i;
	}

int8_t r=decode_can_message(m, val);
output_can_values(val,r,out);
}

void output_can_values(CAN_VALUE *val, int8_t n, FILE *out)
{
int8_t  i;
for(i=0;i<n; ++i) output_can_value(val+i,out);
}


void output_can_value(CAN_VALUE *val, FILE *out)
{
char line[256];
sprintf_can_name_value(val,line);
fputs(line,out);
fputs("\n",out);
}


void print_can_message(CAN_MESSAGE *m, FILE *out, uint8_t obd2only)
{
if(obd2only && (m->id < 0x7e8 || m->id >0x7ef)) return;

fprintf(out, "CAN %3X %d ",m->id,m->length);
int i;
for(i=0; i<m->length; ++i) fprintf(out,"%02X ",m->data.byte[i]);
fputs("\n",out);
}


void create_OBD2_query(uint8_t service, uint16_t pid, CAN_MESSAGE *m)
{
m->id = 0x7df;
m->data.byte[1]=service;
m->length=8;
if(service>15) //custom service
	{
	m->data.byte[0]=3;
	m->data.byte[2]=pid>>8; //big endian
	m->data.byte[3]=pid&0xff;
	}
else
	{
	m->data.byte[0]=2;
	m->data.byte[2]=pid;
	m->data.byte[3]=0x55;
	}
m->data.byte[4]=0x55;
m->data.byte[5]=0x55;
m->data.byte[6]=0x55;
m->data.byte[7]=0x55;
}


//this is a simple general routine, but pid-specific parsing is done via automatically generated code from obd2.dbc
//return nonzero when reply is unsupported pid or other inconsistencies
uint8_t parse_OBD2_reply(const CAN_MESSAGE *m, uint16_t *address, uint8_t *service, uint16_t *pid, uint8_t *abcd)
{
if(m->length!=1+m->data.byte[0]) return 1;
*address=m->id;
if(m->data.byte[1] == 0x7f)
	{
	if(m->length!=4) return 2;
	if(m->data.byte[3]!=0x31) return 3;
	return 4;
	}
if((m->data.byte[1]&0x40) == 0) return 5;
*service = m->data.byte[1]-0x40;
uint8_t data_at;
if(*service>15)  //custom service
        {
	*pid =  m->data.byte[2]<<8; //big endian
	*pid |= m->data.byte[3];
	data_at=4;
	}
else //standard service
	{
	*pid = m->data.byte[2];
	data_at=3;
	}
abcd[0] =  m->data.byte[data_at++];
abcd[1] =  m->data.byte[data_at++];
abcd[2] =  m->data.byte[data_at++];
abcd[3] =  m->data.byte[data_at++];
return 0;
}

//v must be reply to a query on a particular pid bank
int8_t S1pid_is_unsupported(CAN_VALUE *v, uint8_t pid)
{
if((pid&0xe0)!=v->pid) return -1; //wrong reply pid
return (v->value.uint32 &  ((((uint32_t)1) << (31 - ((pid-1)&31))))) ? 0 : 1;
}

