/*
 *
 * Ported to STM32F37, direct connection to GPIO port D; (c) 2019 Jiri Pittner <jiri@pittnerovi.com>
 * Code contributed under the GPL v3 license or higher
 *
 *
*/

#include <stdio.h>
#include "T6963_STM32portD.h"

//STM32 hardware specific
#include "stm32f37x_rcc.h"
#include "stm32f37x_pwr.h"
#include "stm32f37x_tim.h"
#include "stm32f37x_gpio.h"


static inline  void bitWrite(uint16_t &gpioData, int bit, int value)
{
if(value) gpioData |= ((uint16_t)1 << bit);
else gpioData &= ~((uint16_t)1 << bit);
}


T6963_STM32::T6963_STM32()
{
//internal variables init
	_backlight = T6963_BLK_DEFAULT;
	_hrdwFS = T6963_FS_DEFAULT;
	_hrdwRV = T6963_SREVERSE_DEFAULT;
	_fastMode = T6963_FASTMODE_DEFAULT;
}

void T6963_STM32::setBacklight(uint8_t value){
	if (value > 1) value = 1;
	_backlight = value;
	if(_backlight) sbi(GPIOD,GPIO__BLK); else cbi(GPIOD,GPIO__BLK);
}

void T6963_STM32::hardwareFontSelect(bool val){
	_hrdwFS = val;	
	if(_hrdwFS) sbi(GPIOD,GPIO__FS); else cbi(GPIOD,GPIO__FS);
}

//true:reversed,false:normal
void T6963_STM32::hardwareScreenReverse(bool val){
	_hrdwRV = !val;
	if(_hrdwRV)  sbi(GPIOD,GPIO__RV); else cbi(GPIOD,GPIO__RV);
}

void T6963_STM32::chipWaitState(){
	if (_fastMode){
		T6963_WAIT_DELAY();
	} else {
		while(!(checkState() & 0x03));
	}
}


void T6963_STM32::hardwareReset(){
/*
	uint16_t gpioData = 0xFFFF;//start with this
	bitWrite(gpioData,GPIO__RST,0);//Hardware font select
	GPIO_Write(GPIOD,gpioData);
	T6963_WAIT_DELAY();
	bitWrite(gpioData,GPIO__RST,1);//Hardware font select
	GPIO_Write(GPIOD,gpioData);
	T6963_WAIT_DELAY();
*/
	setTextHome(0);
	setGraphicHome(0);
	//Set Graphics Home Address
	sendData(_graphicHome & 0xFF);
	T6963_WAIT_DELAY();
	sendData(_graphicHome >> 8);
	T6963_WAIT_DELAY();
	sendCommand(T6963_SET_GRAPHIC_HOME_ADDRESS);
	//Set Graphics Area
	sendData(_graphicArea);
	T6963_WAIT_DELAY();
	sendData(0x00);
	T6963_WAIT_DELAY();
	sendCommand(T6963_SET_GRAPHIC_AREA);
	T6963_WAIT_DELAY();
	//Set Text home address
	sendData(_textHome & 0xFF);
	T6963_WAIT_DELAY();
	sendData(_textHome >> 8);
	T6963_WAIT_DELAY();
	sendCommand(T6963_SET_TEXT_HOME_ADDRESS);
	T6963_WAIT_DELAY();
	//Set Text Area
	sendData(_graphicArea);
	T6963_WAIT_DELAY();
	sendData(0x00);
	T6963_WAIT_DELAY();
	sendCommand(T6963_SET_TEXT_AREA);
	T6963_WAIT_DELAY();
	//Set Internal CGRAM address
	sendData(((_memSize/2)-1));
	T6963_WAIT_DELAY();
	sendData(0x00);
	T6963_WAIT_DELAY();
	sendCommand(T6963_SET_OFFSET_REGISTER);
	T6963_WAIT_DELAY();
	setDispMode(TEXT_ON,GRAPHIC_ON,CURSOR_OFF,CURSOR_NORMAL);
	T6963_WAIT_DELAY();
	setMode(NORMAL,INT);
	T6963_WAIT_DELAY();
	clearText();
	clearGraphic();
	T6963_WAIT_DELAY();
	T6963_WAIT_DELAY();
	T6963_WAIT_DELAY();
}

void T6963_STM32::begin(uint8_t width, uint8_t height, uint8_t charsize,uint8_t memsize)
{

//internal variables init
        _width = width;                                                 //240
        _height = height;                                               //128
        _fontWidth = charsize;                                  //6 or 8
        if(memsize>64) memsize=64;
        _memSize = memsize;             //32
        _graphicArea = _width / _fontWidth;             //40    (6)
        _graphicSize = _graphicArea * _width;   //9600  (6)
        //_textSize = _graphicSize / 8;                 //1200  (6) maybe _graphicSize / _fontWidth;
        _textSize = _graphicSize / _fontWidth;
        _maxCol = _graphicArea - 1;                             //39    (6)
        //_maxRow = (_width / 8) - 1;                   //29    (6) maybe (_width / _fontWidth) - 1;
        _maxRow = (_width / _fontWidth) - 1;


//GPIO init
{
GPIO_InitTypeDef GPIO_InitStructure;
RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOD, ENABLE);
GPIO_InitStructure.GPIO_Pin = 0xffff;
GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
GPIO_InitStructure.GPIO_Speed = GPIO_Speed_10MHz;
GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
GPIO_Init(GPIOD, &GPIO_InitStructure);
GPIO_Write(GPIOD,0xffff);
}

// GPIO_Speed_2MHz does not help occasional miscommunication


//display reset
	hardwareReset();
}


void T6963_STM32::sendCommand(byte command){

	chipWaitState();
	uint16_t gpioData = (0xFFFF << 8) | command;
	bitWrite(gpioData,GPIO__FS,_hrdwFS);//Hardware font select
	bitWrite(gpioData,GPIO__RV,_hrdwRV);//Hardware screen reverse
	bitWrite(gpioData,GPIO__BLK,_backlight);//backlight
	gpioData &= ~((1 << GPIO__WR) | (1 << GPIO__CE));//clear
	GPIO_Write(GPIOD,gpioData);
        T6963_WAIT_DELAY();
	gpioData |= ((1 << GPIO__WR) | (1 << GPIO__CE)); //setta
	GPIO_Write(GPIOD,gpioData);
        T6963_WAIT_DELAY();
}

void T6963_STM32::sendData(byte data){
	chipWaitState();
	uint16_t gpioData = (0xFFFF << 8) | data;
	bitWrite(gpioData,GPIO__FS,_hrdwFS);//Hardware font select
	bitWrite(gpioData,GPIO__RV,_hrdwRV);//Hardware screen reverse
	bitWrite(gpioData,GPIO__BLK,_backlight);//backlight
	gpioData &= ~((1 << GPIO__CD) | (1 << GPIO__WR) | (1 << GPIO__CE));//clear
//fprintf(stderr,"write1 %x\n",gpioData);
	GPIO_Write(GPIOD,gpioData);
	T6963_WAIT_DELAY();
	gpioData |= ((1 << GPIO__CD) | (1 << GPIO__WR) | (1 << GPIO__CE)); //setta
//fprintf(stderr,"write2 %x\n",gpioData);
	GPIO_Write(GPIOD,gpioData);
	T6963_WAIT_DELAY();
}

uint8_t T6963_STM32::readData(bool do_CD){
	uint8_t tmp = 0xFF;
	uint16_t gpioData = 0xFFFF;//start with this
	chipWaitState();
{
GPIO_InitTypeDef GPIO_InitStructure;
GPIO_InitStructure.GPIO_Pin = 0x00ff;
GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN;
GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
GPIO_Init(GPIOD, &GPIO_InitStructure);
}
	bitWrite(gpioData,GPIO__FS,_hrdwFS);//Hardware font select
	bitWrite(gpioData,GPIO__RV,_hrdwRV);//Hardware screen reverse
	bitWrite(gpioData,GPIO__BLK,_backlight);//backlight
	gpioData &= ~((do_CD?(1 << GPIO__CD):0) | (1 << GPIO__RD) | (1 << GPIO__CE));//clear
	GPIO_Write(GPIOD,gpioData);
        T6963_WAIT_DELAY();
	uint16_t p = GPIO_ReadInputData(GPIOD);
	tmp = p & 0xFF;//get lower 8 bits
{
GPIO_InitTypeDef GPIO_InitStructure;
GPIO_InitStructure.GPIO_Pin = 0x00ff;
GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
GPIO_InitStructure.GPIO_Speed = GPIO_Speed_10MHz;
GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
GPIO_Init(GPIOD, &GPIO_InitStructure);
}
	gpioData |= ((do_CD?(1 << GPIO__CD):0) | (1 << GPIO__RD) | (1 << GPIO__CE)); //setta
	GPIO_Write(GPIOD,gpioData);
        T6963_WAIT_DELAY();
	return tmp;
}

uint8_t T6963_STM32::checkState(void){
	return readData(0);
}

