/*
 * lan.c  - higher level ethernet routines based on etherShield
 *  
 *        Copyright (C) 2019-2022 Jiri Pittner <jiri@pittnerovi.com>
 *
 *         This program is free software: you can redistribute it and/or modify
 *         it under the terms of the GNU General Public License as published by
 *         the Free Software Foundation, either version 3 of the License, or
 *         (at your option) any later version.
 *       
 *          This program is distributed in the hope that it will be useful,
 *          but WITHOUT ANY WARRANTY; without even the implied warranty of
 *          MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *          GNU General Public License for more details.
 *          You should have received a copy of the GNU General Public License
 *          along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *          
 *        
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "usart123.h"
#include "delay.h"
#include "stm32f37x_gpio.h"
#include "stm32f37x_rcc.h" 
#include "stm32f37x_pwr.h"
#include "stm32f37x_tim.h"
#include "stm32f37x_exti.h"
#include "stm32f37x_misc.h"
#include "stm32f37x_syscfg.h"

#include "lan.h"

#define printP puts
#define print puts
#define PSTR(x) (x)

uint8_t dnsip[4] = {0,0,0,0};

volatile uint8_t inudplen=0;
volatile uint8_t inudpline=0;
volatile char inudp[INUDPMAX+1];


uint8_t init_lan(uint8_t newdhcp)
{
watchdog();
        /*initialize enc28j60*/
        enc28j60Init(mymac);
        //enc28j60clkout(2); // we do not use clockout 
        delay_ms(10);
        
	/* Magjack leds configuration, see enc28j60 datasheet, page 11 */
	// LEDA=greed LEDB=yellow
	//
	// 0x880 is PHLCON LEDB=on, LEDA=on
	// enc28j60PhyWrite(PHLCON,0b0000 1000 1000 00 00);
	enc28j60PhyWrite(PHLCON,0x880);
	delay_ms(250); delay_ms(250);
	//
	// 0x990 is PHLCON LEDB=off, LEDA=off
	// enc28j60PhyWrite(PHLCON,0b0000 1001 1001 00 00);
	enc28j60PhyWrite(PHLCON,0x990);
	delay_ms(250); delay_ms(250);
	//
	// 0x880 is PHLCON LEDB=on, LEDA=on
	// enc28j60PhyWrite(PHLCON,0b0000 1000 1000 00 00);
	enc28j60PhyWrite(PHLCON,0x880);
	delay_ms(250); delay_ms(250);
	//
	// 0x990 is PHLCON LEDB=off, LEDA=off
	// enc28j60PhyWrite(PHLCON,0b0000 1001 1001 00 00);
	enc28j60PhyWrite(PHLCON,0x990);
	delay_ms(250); delay_ms(250);
	//
        // 0x476 is PHLCON LEDA=links status, LEDB=receive/transmit
        // enc28j60PhyWrite(PHLCON,0b0000 0100 0111 01 10);
        enc28j60PhyWrite(PHLCON,0x476);
	delay_ms(100);

watchdog(); 
        //init the ethernet/ip layer:
        if(myip[0]!=0 && !newdhcp)
		{
		printP(PSTR("Skipping DHCP...\n"));
		 init_ip_arp_udp_tcp(mymac,myip,80);
		return 0;
		}
	else	
		{
		printP(PSTR("Trying DHCP...\n"));
		uint8_t r=allocateIPAddress(lan_buf, LAN_BUFFER_SIZE, &mymac, src_port, &myip, &mynetmask, &mygwip, &dnsip, &dhcpsrvip );
		if(r==0) 
			{
			printP(PSTR("DHCP timed out\n"));
			return 1;
			}
		else 
			{
			char c[48];
			sprintf(c,"IP %d.%d.%d.%d\n",myip[0],myip[1],myip[2],myip[3]);
			print(c);
			sprintf(c,"MSK %d.%d.%d.%d\n",mynetmask[0],mynetmask[1],mynetmask[2],mynetmask[3]);
                        print(c);
			sprintf(c,"GW %d.%d.%d.%d\n",mygwip[0],mygwip[1],mygwip[2],mygwip[3]);
                        print(c);
			sprintf(c,"DNS %d.%d.%d.%d\n",dnsip[0],dnsip[1],dnsip[2],dnsip[3]);
                        print(c);
			printP(PSTR("DHCP successfull\n"));
			//GW MAC not set yet at this point
			return 0;
			}
		}
return 1;
}





uint8_t tcpfinished=0;

enum smtp {start=0,connected, helo, mailfrom, rcptto, data, closed};
static volatile enum smtp session=0;
static char *smtp_from, *smtp_to, *smtp_text;
static volatile uint8_t smtp_failed;

uint8_t smtp_client_tcp_result_callback(uint8_t fd, uint8_t statuscode,uint16_t data_start_pos_in_buf, uint16_t len_of_data)
{
watchdog();
uint8_t close_tcp_session=0;
if(!statuscode)
	{
	if(len_of_data==0) {printP(PSTR("NOTE: empty packet received\n")); return(close_tcp_session);}
	char *msg;
	msg = (char *) &lan_buf[data_start_pos_in_buf];
	uint8_t z;
	z=lan_buf[data_start_pos_in_buf+len_of_data];
	lan_buf[data_start_pos_in_buf+len_of_data]=0;
	print(msg);
	lan_buf[data_start_pos_in_buf+len_of_data]=z;
uint16_t code = atoi(msg);
{
char z[32];
sprintf(z,"code %d\n",code);
print(z);
}
close_tcp_session = 0xff; //indicate sending reply
switch(session){
	case start: //check initial message
		if(code==220) session=connected; else close_tcp_session=1;
		break;
	case connected: //check helo
		if(code==250) session=helo; else close_tcp_session=1;
		break;
	case helo: //check mailfrom
		if(code==250) session=mailfrom; else close_tcp_session=1;
		break;
	case mailfrom: //check rcptto
		if(code==250) session=rcptto; else close_tcp_session=1;
		break;
	case rcptto: //check 354
		if(code==354) session=data; else close_tcp_session=1;
		break;
	case data: //check 250
		if(code==250) session=closed; else close_tcp_session=1;
		break;
	case closed: //check 221 bye and close anyway
		if(code==221) smtp_failed=0;
		close_tcp_session=1;
		break;
}
	}
else
	{
	printP(PSTR("TCP receive error\n"));
	close_tcp_session=1;
	}
tcpfinished= (close_tcp_session==1);
if(tcpfinished) printP(PSTR("closing session\n"));
return(close_tcp_session);
}


uint16_t smtp_client_tcp_datafill_callback(uint8_t fd)
{
watchdog();
int16_t len_of_data_filled_in = 0;
char msg[128];
switch(session){
        case start:
		msg[0]=0;
                break;
        case connected: //send helo
		{
		strcpy(msg,"helo noreply.pittnerovi.com\r\n");
		len_of_data_filled_in=  fill_tcp_data(lan_buf,len_of_data_filled_in, msg);
		}
		break;
	case helo: //send mailfrom
		{
                strcpy(msg,"mail from: ");
                strcat(msg,smtp_from);
                strcat(msg,"\r\n");
                len_of_data_filled_in=  fill_tcp_data(lan_buf,len_of_data_filled_in, msg);
                }
                break;
        case mailfrom: //send rcptto
                {
                strcpy(msg,"rcpt to: ");
                strcat(msg,smtp_to);
                strcat(msg,"\r\n");
                len_of_data_filled_in=  fill_tcp_data(lan_buf,len_of_data_filled_in, msg);
                }
                break;
        case rcptto: //send data 
		strcpy(msg,"data\r\n");
		len_of_data_filled_in=  fill_tcp_data(lan_buf,len_of_data_filled_in, msg);
                break;
        case data: //send text
		{
		len_of_data_filled_in= fill_tcp_data(lan_buf,len_of_data_filled_in, smtp_text);
		print(smtp_text);
		strcpy(msg,"\r\n.\r\n");
                len_of_data_filled_in=  fill_tcp_data(lan_buf,len_of_data_filled_in, msg);
		}
                break;
        case closed: //send quit
		strcpy(msg,"quit\r\n");
                len_of_data_filled_in=  fill_tcp_data(lan_buf,len_of_data_filled_in, msg);
		break;
}
print(msg);
return(len_of_data_filled_in);
}



uint8_t smtp_session(uint8_t *destip_in, char *from, char *to, char *text) //destip
{
smtp_failed=1;
watchdog();
smtp_from=from;
smtp_to=to;
smtp_text=text;
memcpy(destip,destip_in,4);
tcpfinished=0;
client_tcp_set_serverip(destip);
session=start;
printP(PSTR("Initiating SMTP session\n"));
uint8_t fd=client_tcp_req(&smtp_client_tcp_result_callback,&smtp_client_tcp_datafill_callback,25,0);

uint32_t timeout=1000000;
while(!tcpfinished)
	{
	watchdog();
	uint16_t plen = enc28j60PacketReceive(LAN_BUFFER_SIZE, lan_buf);
	watchdog();
	uint16_t dat_p=packetloop_icmp_tcp(lan_buf, plen);
	if(--timeout == 0) return 2;
	}
return smtp_failed;
}


void ping_send(uint8_t *dest)
{
memcpy(destip,dest,4);
client_set_gwip(dest); //send directly in lan, not via router
reset_arp_state();
memset(lan_buf,0x00,LAN_BUFFER_SIZE); //cleanup
client_icmp_request(lan_buf,dest);
}


void udp_send(uint8_t *dest,uint16_t dst_port,uint16_t src_port,uint8_t *msg, uint16_t len)
{
watchdog();
if(!msg) return;
if(len==0) return;

memcpy(destip,dest,4);
client_set_gwip(dest); //send directly in lan, not via router
reset_arp_state();


uint16_t ck;
memset(lan_buf,0x00,LAN_BUFFER_SIZE); //cleanup
send_udp_prepare(lan_buf,src_port,dest,dst_port);
memcpy(&lan_buf[ETH_SRC_MAC],mymac,6);	
lan_buf[ETH_TYPE_H_P]=ETHTYPE_IP_H_V;
lan_buf[ETH_TYPE_L_P]=ETHTYPE_IP_L_V;
memcpy(&lan_buf[IP_SRC_P],myip,4);
memcpy(&lan_buf[IP_DST_P],dest,4);
lan_buf[IP_HEADER_LEN_VER_P]=0x45;
lan_buf[IP_PROTO_P]=IP_PROTO_UDP_V;

// total length field in the IP header must be set:
if(len>LAN_BUFFER_SIZE-UDP_HEADER_LEN-1) len = LAN_BUFFER_SIZE-UDP_HEADER_LEN-1;
lan_buf[IP_TOTLEN_H_P]=0;
lan_buf[IP_TOTLEN_L_P]=IP_HEADER_LEN+UDP_HEADER_LEN+len;
fill_ip_hdr_checksum(lan_buf);

lan_buf[UDP_DST_PORT_H_P]=dst_port>>8;
lan_buf[UDP_DST_PORT_L_P]=dst_port & 0xff;
lan_buf[UDP_SRC_PORT_H_P]=src_port>>8;	
lan_buf[UDP_SRC_PORT_L_P]=src_port & 0xff;
// calculte the udp length:
lan_buf[UDP_LEN_H_P]=0;
lan_buf[UDP_LEN_L_P]=UDP_HEADER_LEN+len;
// zero the checksum
lan_buf[UDP_CHECKSUM_H_P]=0;
lan_buf[UDP_CHECKSUM_L_P]=0;
memcpy(&lan_buf[UDP_DATA_P],msg,len);
ck=checksum(&lan_buf[IP_SRC_P], 16 + len,1);
lan_buf[UDP_CHECKSUM_H_P]=ck>>8;
lan_buf[UDP_CHECKSUM_L_P]=ck& 0xff;
enc28j60PacketSend(UDP_HEADER_LEN+IP_HEADER_LEN+ETH_HEADER_LEN+len,lan_buf);
}


uint8_t dns_session(char *domain, uint8_t *mailserver, uint8_t mx)
{
uint32_t ii;
char *mxhost;
char *p=domain;
while(*p) {if(*p=='\n') *p=0; else ++p;}

if(mx) 
{
char mxhostbuf[64];
watchdog();                                                                                                                   memset(lan_buf,0,LAN_BUFFER_SIZE); //some iniialization is missing in dnslkup_request                                             dnslkup_request(lan_buf,domain,15); // mx lookup
for(ii=0; ii<100000; ++ii)
        {
        watchdog();
        uint16_t plen = enc28j60PacketReceive(LAN_BUFFER_SIZE, lan_buf);
        watchdog();
        uint16_t dat_p=packetloop_icmp_tcp(lan_buf, plen);
        watchdog();
        uint8_t r=udp_client_check_for_dns_answer(lan_buf,plen,mxhostbuf);
        if(r)
                {
                printf("MX DNS lookup succesfull: %s\n",mxhostbuf);
                strcat(mxhostbuf,".");
                strcat(mxhostbuf,domain);
                mxhost=mxhostbuf;
                memset(lan_buf,0,LAN_BUFFER_SIZE); //some iniialization is missing in dnslkup_request
                goto hostlookup;
                }
        }
puts("MX DNS lookup failed\n");
return 2;
} 
else
mxhost=domain;          
  
hostlookup:
printf("LOOKING UP %s\n",mxhost);
watchdog();
dnslkup_request(lan_buf,mxhost,1); // host lookup
for(ii=0; ii<100000; ++ii)
        {
        watchdog();
        uint16_t plen = enc28j60PacketReceive(LAN_BUFFER_SIZE, lan_buf);
        watchdog();
        uint16_t dat_p=packetloop_icmp_tcp(lan_buf, plen);
        watchdog();
        uint8_t r=udp_client_check_for_dns_answer(lan_buf,plen,NULL);
        if(r)
                {
                memcpy(mailserver,dnslkup_getip(),4);
                printf("HOST DNS lookup succesfull: %d.%d.%d.%d\n",mailserver[0],mailserver[1],mailserver[2],mailserver[3]);
                return 0;
                }
        }
puts("HOST DNS lookup failed\n");
return 1; //no success
}

uint8_t dns_smtp_session(char *from, char *to, char *text)
{
//we go via the gateway as the mailserver can be outside lan
//gateway is supposed to do hairpinning if it is inside

client_set_gwip(mygwip); //send directly in lan, not via router
reset_arp_state();


uint8_t mailserver[4];
char *domain=strchr(to,'@')+1;
uint8_t r=dns_session(domain,mailserver,1);
if(r) //try backup dns
        {
        dnslkup_set_dnsip(dnsip2);
        r=dns_session(domain,mailserver,1);
        }
if(r) return 10*r;
return smtp_session(mailserver, from, to, text);
}


void write_udp2(char *p)
{
if(!p) return;
if(p[0]==0) return;
write_udp(p,strlen(p));
}


void write_udp(uint8_t *msg, uint8_t datalen) 
{

			{
		        uint16_t ck;
		memset(lan_buf,0x00,LAN_BUFFER_SIZE); //cleanup
		memcpy(&lan_buf[ETH_DST_MAC],dstmac,6);	
		memcpy(&lan_buf[ETH_SRC_MAC],mymac,6);	
		lan_buf[ETH_TYPE_H_P]=ETHTYPE_IP_H_V;
		lan_buf[ETH_TYPE_L_P]=ETHTYPE_IP_L_V;
		memcpy(&lan_buf[IP_SRC_P],myip,4);
		memcpy(&lan_buf[IP_DST_P],destip,4);
		lan_buf[IP_HEADER_LEN_VER_P]=0x45;
		lan_buf[IP_PROTO_P]=IP_PROTO_UDP_V;

			// total length field in the IP header must be set:
			if(datalen>LAN_BUFFER_SIZE-UDP_HEADER_LEN-1) datalen = LAN_BUFFER_SIZE-UDP_HEADER_LEN-1;
			lan_buf[IP_TOTLEN_H_P]=0;
		        lan_buf[IP_TOTLEN_L_P]=IP_HEADER_LEN+UDP_HEADER_LEN+datalen;
		fill_ip_hdr_checksum(lan_buf);

			lan_buf[UDP_DST_PORT_H_P]=dst_port>>8;
			lan_buf[UDP_DST_PORT_L_P]=dst_port & 0xff;
			lan_buf[UDP_SRC_PORT_H_P]=src_port>>8;	
			lan_buf[UDP_SRC_PORT_L_P]=src_port & 0xff;
			// calculte the udp length:
			lan_buf[UDP_LEN_H_P]=0;
        		lan_buf[UDP_LEN_L_P]=UDP_HEADER_LEN+datalen;
			// zero the checksum
			lan_buf[UDP_CHECKSUM_H_P]=0;
        		lan_buf[UDP_CHECKSUM_L_P]=0;
			strncpy(&lan_buf[UDP_DATA_P],msg,datalen);
			ck=checksum(&lan_buf[IP_SRC_P], 16 + datalen,1);
        		lan_buf[UDP_CHECKSUM_H_P]=ck>>8;
        		lan_buf[UDP_CHECKSUM_L_P]=ck& 0xff;
        		enc28j60PacketSend(UDP_HEADER_LEN+IP_HEADER_LEN+ETH_HEADER_LEN+datalen,lan_buf);
			}

}


char *read_udp(uint8_t n_white_ip, uint8_t ip_whitelist[][4]) //returns NULL or pointer to payload
{

		memset(lan_buf,0x00,LAN_BUFFER_SIZE); //cleanup


                // get the next new packet:
                int plen = enc28j60PacketReceive(LAN_BUFFER_SIZE, lan_buf);

                /*plen will be unequal to zero if there is a valid 
                 * packet (without crc error) */
                if(plen==0){
                        return NULL;
                }

		//printf("Packet received\n");

                // arp is broadcast if unknown but a host may also // verify the mac address by sending it to // a unicast address.
                if(eth_type_is_arp_and_my_ip(lan_buf,plen))
                        {
                        make_arp_answer_from_request(lan_buf);
                        return NULL;
                        }

                // check if ip packets (icmp or udp) are for us:
                if(eth_type_is_ip_and_my_ip(lan_buf,plen)==0) return NULL;
                        
                if(lan_buf[IP_PROTO_P]==IP_PROTO_ICMP_V && lan_buf[ICMP_TYPE_P]==ICMP_TYPE_ECHOREQUEST_V){
                        // a ping packet, let's send pong
                        make_echo_reply_from_request(lan_buf,plen);
                        return NULL;
                        }

		//check against the whitelist
		int i;
		for(i=0; i<n_white_ip; ++i)
			{
			int r=memcmp(&lan_buf[IP_SRC_P],&ip_whitelist[i][0],4);
			//printf("test %d: %d %d %d %d %d %d %d %d\n",r,lan_buf[IP_SRC_P],lan_buf[IP_SRC_P+1],lan_buf[IP_SRC_P+2],lan_buf[IP_SRC_P+3],ip_whitelist[i][0],ip_whitelist[i][1],ip_whitelist[i][2],ip_whitelist[i][3]);
			if(!r) 
				{
				//fprintf(stderr,"UDPin accepted\n");
				goto process_udp;
				}
			}
		//fprintf(stderr,"UDPin rejected\n"); 
		return 0;

		
                // udp incoming
		process_udp:
                if (lan_buf[IP_PROTO_P]==IP_PROTO_UDP_V)
			{
			//set reply destination
			memcpy(dstmac,&lan_buf[ETH_SRC_MAC],6);
			memcpy(destip,&lan_buf[IP_SRC_P],4);
			//handle payload (limited size)
                        uint16_t payloadlen=lan_buf[UDP_LEN_L_P]-UDP_HEADER_LEN;
                        lan_buf[UDP_DATA_P+payloadlen]=0;
			if(payloadlen>INUDPMAX) payloadlen=INUDPMAX;
			memcpy(inudp,lan_buf+UDP_DATA_P,payloadlen+1);
			inudpline=inudplen=payloadlen;
			return lan_buf+UDP_DATA_P;
                	}
return 0;
}


