/*
    Copyright (C) 2021 Jiri Pittner <jiri@pittnerovi.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include <stdio.h>
#include <math.h>
#include "vecmat3.h"
#include "quaternion.h"
#include "simple.h"
#include "avi_calibration.h"
#include "magnetometer.h"

using namespace LA_Quaternion;
using namespace LA_Vecmat3;
using namespace LA_Simple;

//magnetometer calibration routines, based on the ST design tip DT0059 by Andrea Vitali (2018)

//presently we employ the non-rotated ellipsoid model for simplicity and numerical stability
Vec3<float> magnetic_offsets;
Vec3<float> magnetic_scales;
simple_linfit<float,FITVARS> magnetic_fit;

//initialize the fit
void init_magnetic_calibration(void)
{
magnetic_fit.clear(false);
}

//accumulate data for the fit
void accumulate_magnetic_calibration(const Vec3<float> &m,const int periodic_update, int eeprom_ok, const float preserve)
{
float funcs[FITVARS];
for(int i=0; i<3; ++i)
	{
	funcs[i] = m[i]*m[i];
	funcs[3+i] = 2*m[i];
	}
magnetic_fit.input(funcs,1);
if(periodic_update && magnetic_fit.npoints>=periodic_update)
	{
	fit_magnetic_calibration(eeprom_ok,preserve);
	magnetic_fit.npoints=0; //count points again
	}
}

//perform the fit and store results to eeprom
void fit_magnetic_calibration(int eeprom_ok, float preserve)
{
/*
printf("solving from %d points\n",magnetic_fit.npoints);
for(int i=0; i<FITVARS; ++i)
	{
	for(int j=0; j<FITVARS; ++j) printf("%f ",magnetic_fit.fitmat[i][j]);
	printf("\n");
	}
*/

float det=magnetic_fit.solve(preserve); //preserve data scaled down
/*
printf("after\n");
for(int i=0; i<FITVARS; ++i)
        {
        for(int j=0; j<FITVARS; ++j) printf("%f ",magnetic_fit.fitmat[i][j]);
        printf("\n");
        }
*/

if(abs(det)<1e-7) {printf("magnetic_fit determinant too small %g\n",det); return;}

printf("mfit det %g, preservation %f\n",det,preserve);
printf("mfit abc %f %f %f\n",magnetic_fit[0],magnetic_fit[1],magnetic_fit[2]);
printf("mfit ghi %f %f %f\n",magnetic_fit[3],magnetic_fit[4],magnetic_fit[5]);

//compute offsets and scales from the fit
//these scales will cause magnetic vector to be rescaled to unit sphere

float G=1;
for(int i=0; i<3; ++i) 
	{
	magnetic_offsets[i]=magnetic_fit[3+i]/magnetic_fit[i];
	G += magnetic_fit[3+i]* magnetic_offsets[i];
	}
for(int i=0; i<3; ++i) 
	{
	float tmp=magnetic_fit[i]/G;
	if(tmp<0) {printf("negative sqrt in mfit %f\n",tmp);  for(int j=0; j<3; ++j) magnetic_scales[j]=1; break;} //if one is wrong, all considered unreliable
	else magnetic_scales[i] = sqrt(tmp);
	}

//sanitize the results if the fit goes weird
for(int i=0; i<3; ++i) if(abs(magnetic_offsets[i]) > 1.f) magnetic_offsets[i]=0.f;
for(int i=0; i<3; ++i)
        {
	if(abs(magnetic_scales[i]) < 0.2f || magnetic_scales[i] > 5.f) {for(int j=0; j<3; ++j) magnetic_scales[j]=1; break;} //if one is wrong, all considered unreliable
	}

printf("mfit offsets %f %f %f\n",magnetic_offsets[0],magnetic_offsets[1],magnetic_offsets[2]);
printf("mfit scales %f %f %f\n",magnetic_scales[0],magnetic_scales[1],magnetic_scales[2]);


store_magnetic_calibration(eeprom_ok,0,magnetic_offsets,magnetic_scales,&magnetic_fit.fitmat[0][0],&magnetic_fit.rhsmat[0][0]);
}

//reset the calibration
void reset_magnetic_calibration(int eeprom_ok)
{
store_magnetic_calibration(eeprom_ok,1,magnetic_offsets,magnetic_scales,&magnetic_fit.fitmat[0][0],&magnetic_fit.rhsmat[0][0]);
}

//apply correction
Vec3<float> magnetometer_corrected(const Vec3<float> &mraw, bool zero_z)
{
Vec3<float> r;
r= magnetic_scales.elementwise_product(mraw+magnetic_offsets);
//as the calibration is presumably done on ground or in level flight, the inclination component would be just subtracted anyway as a constant offset
//so it is probably better to zero it, and set it to zero also in the magnetic reference vector for AHRS
//so that magnetometer shall contribute to determination of YAW only
if(zero_z) r[2]=0;
return r;
}

