/*
 * *      milliohm.c  ... firmware for a milliohm/ESR meter based on STM32F373CCT6 and AD8253
 * *
 * *      Copyright (C) 2012 Jiri Pittner <jiri@pittnerovi.com>
 * *
 * *      This program is free software: you can redistribute it and/or modify
 * *      it under the terms of the GNU General Public License as published by
 * *      the Free Software Foundation, either version 3 of the License, or
 * *      (at your option) any later version.
 * *
 * *      This program is distributed in the hope that it will be useful,
 * *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 * *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * *      GNU General Public License for more details.
 * *      You should have received a copy of the GNU General Public License
 * *      along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * *
 * *
 * */


/*TODO:
 * implement the analogue lock-in amplifier and integrator way of measurement
 * implement another way to measure capacitance using DAC to create a reference level and comparator interrupt to measure time to charge
 */

/* QUICK USER MANUAL
0) Discharge a capacitor before testing. Connect the DUT in the 4-leads (Kelvin) setup.
   Notice that with disconnected probes, some bogus resistance is displayed due to current flowing through the safety diodes.
1) Press '0' repeatedly to select the component type - rotates among R,C,L,menu. Only DC resistance of an inductor is measured.
   In resistor mode, voltage/current is displayed at second line. 
2) Press '#' and '1'-'4' to select test current (1=0.0005A, 2=0.005A, 3=0.05A, 4=0.5A).
   Test current selection is NOT INCLUDED IN AUTORANGING for safety - high test currents might destroy some devices.
   During autoranging, the '!'  sign indicates sub-optimal accuracy of the result, after successfull autoranging 
   it is replaced by '=' sign. Note that for certain combinations of component value and test current the autoranging cannot succeed.
3) Optionally press '*xy' to manually select gains of AD8253 (x=0-3) and SDADC (y=0-4)
4) To go back to autoranging, press '**'
5) For capacitor, optionally press '*#' to enter capacitance measurement mode, press '**' to go back to ESR mode
   Notice that for accurate measurement of capacitance you should select the lowest current, definitely '#1' for capacitors below 10mF.
   On the other hand, for ESR measurement the test current should be proportional to the capacitance of the capacitor using the   lowest current for capacitances below 100uF.
   Optionally press '##nnn' to enter charging target voltage in mV (<400mV) for capacitance measurement.
6) Display readout should be self-explanatory, consult source code for details. Right upper corner shows 
   measurement type, test current, autoranging mode, and selected voltage gains.
   To supress noise and improve accuracy, long time averaging is performed, so the display update can take a few seconds.
7) USART (3V3 level) can be used to issue commands and log the results. It is NOT galvanically separated from probe leads.
8) Voltage probes are usable as a separate microvoltmeter with resolution of 0.1uV at the highest gain. 
   Of course, this accuracy can be achieved only in the lock-in setup to cancel noise and offset.
   The current must thus be switched to external sync (PA5) by '#5' and current probes should be shorted to avoid a bias!
   The DUT must use the sync signal to provide the measured voltage only during logical high of the sync signal.
   Display of R is then meaningless while the first number at the second line is the voltage in mV.
 9)Pressing '9' in menu mode toggles display backlight (is not recommended to achieve highest accuracy).
 */


/*hardware: STM32F373CCT6,  XTAL = 7372800ULL , PLLMULT = 10
 * PIN USAGE:
 * 1 VBAT = 3V3digital
 * 2 PC13 = capacitor discharge NPN
 * 3 PC14 = AD8253 gainselect 0
 * 4 PC15 = AD8253 gainselect 1
 * 5 OSC_IN = XTAL
 * 6 OSC_OUT = XTAL
 * 7 NRST = for ISP
 * 8 VSSA = GND analog
 * 9 VDDA = 3V3 analog
 * 10 PA0 = comparator -
 * 11 PA1 = ADC1, comparator +, connected to  AD8253 out, extra buffered
 * 12 PA2 = ADC2, Vbatt sense by voltage divisor
 * 13 PA3 = ADC3, AGND for ADC offset calibration
 * 14 PA4 = DAC reference to comparator -
 * 15 PA5 = ext. sync for microvoltmeter (via 100R)
 * 16 PA6
 * 17 VDD_2  = 3V3 digital
 * 18 PB0 = SDADC1Ch6 Isense_high_med+
 * 19 PB1 = SDADC1Ch6 Isense_high_med-
 * 20 PB2 = SDADC2Ch6 for AD8253 out, extra buffered
 * 21 PE8 = SDADC1Ch8 Isense_low_vlow+
 * 22 PE9 = SDADC1Ch8 Isense_low_vlow-
 * 23 VSS = GND analog
 * 24 VDDSD = 3V3 analog
 * 25 VREFSD = decoupled by ceramic 1uF+100nF
 * 26 PB14
 * 27 PB15 = SDADC3Ch7 = integrator output
 * 28 PD8 = SDADC3Ch6 = AD8253 out via voltage divisor, unbuffered : DO NOT USE, SDADC has too low DC input impedance
 * 29 PA8 = medium current switch PNP (active 0)
 * 30 PA9 = USART1
 * 31 PA10 = USART1
 * 32 PA11 = low current switch (active 1, inactive must be hi-z)
 * 33 PA12 = lowest current switch (active 1, inactive must be hi-z)
 * 34 PA13 = ST-LINK ISP
 * 35 PF6 = lock-in sampling + (hi-z = off, 1=on)
 * 36 PF7 = lock-in sampling - (hi-z = off, 1=on)
 * 37 PA14 = ST-LINK ISP
 * 38 PA15 = high current switch PNP (active 0)
 * 39 PB3 = SCK
 * 40 PB4 = MISO
 * 41 PB5 = MOSI
 * 42 PB6 = CS
 * 43 PB7 = interrupt from B port of MCP23S17
 * 44 BOOT0 = tied to GND
 * 45 PB8 = MCP23S17 reset
 * 46 PB9 = integrator dump (hi-z = off, 1=on)
 * 47 VSS_1 = GND digital
 * 48 VDD_1 = 3V3 digital
 */


/* 
 * BEGIN CALIBRATION INSTRUCTIONS AND RESULTS
 */

//resistors to be measured by ohmmeter when building the device
#define RsenseHM 2.70
#define RsenseLW 269.1
/*
#define RextraH 2.70
#define RextraM 55.9
#define RextraL 268.75
#define RextraW 5605.0
*/
#define Divisor_direct (0.999288*32935./(21890.+32935.))
#define Divisor_integrated (3286./(2209.+3286.))
//due to ADC input impedance, do not use higher resistances than of order of kohms
#define Divisor_battery (1199./(1199.+4683.))
//
//voltages to be measured by a voltmeter after power up at VDDA and VREFSD+ pins, respectively, with respect to VSSA/VREFSD- pin
#define VDD_analog 3.2978
//#define VREFINT1 1.2297
#define VREFINT2 1.8453

//actual voltage at PA2 with respect to ADC reading
#define ADC_factor 1.00272

//shutdown threshold to prevent deep discharge of 6 NiMH cells
#define LOWBAT 7.2 

//for current measurement, compensates inaccuracy of measurement of the sense resistors above
//connect an accurate ampermeter instead of the tested resistor, replace these numbers by 4,.5,4,.5
//switch manually to each test current range, compare the computed value with respect to ampermeter readout
//and put the appropriate factor here to match the results
//
#define NCURRENT 5
const float current_sense_factor[NCURRENT+1] = {1.,4.12183658,0.51355478,4.09562, 0.52191284,1.};

//connect some test resistor and use appropriate testing current to generate a voltage drop suitable for a given gain
//measure at input / output of AD8253 and put the actual gains here

const float ad8253_gain_factors[4] = {0.9995,10.005,100.03,1000.};

//for voltage measurement by sdadc2 (absorbs TL082 non-unity gain factor)
//connect some test resistor and use appropriate testing current
//put the numbers .5,1,2,4,8 here, measure voltage at the resistor and
//compare with the output voltages, put the appropriate factors here to match the results

const float sdadc_gain_factors[5] = {0.5159728,1.03171,2.0658819,4.144030,8.311235};

/*
 * END CALIBRATION
 */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <setjmp.h>

#include "usart.h"
#include "delay.h"
#include "kbd_spi1.h"
#include "lcd_spi1.h"
#include "sdadc.h"
#include "adc.h"
#include "stm32f37x_gpio.h"
#include "stm32f37x_rcc.h"
#include "stm32f37x_pwr.h"
#include "stm32f37x_tim.h"


const uint32_t current_sense_channel[NCURRENT+1] = {SDADC_Channel_8,SDADC_Channel_8,SDADC_Channel_8,SDADC_Channel_6,SDADC_Channel_6,SDADC_Channel_8};
const uint32_t current_sense_gain[NCURRENT+1] = {SDADC_Gain_1,SDADC_Gain_4,SDADC_Gain_1_2,SDADC_Gain_4,SDADC_Gain_1_2,SDADC_Gain_1};
const float current_sense_resistor[NCURRENT+1] = {1e10,RsenseLW,RsenseLW,RsenseHM,RsenseHM,1e10};
const uint32_t sdadc_gain_list[5] = {SDADC_Gain_1_2,SDADC_Gain_1,SDADC_Gain_2,SDADC_Gain_4,SDADC_Gain_8};


float vbattery(void) 
{
init_ADC(ADC_Channel_2);
float v= read_ADC(100);
v *= VDD_analog/4096./100.;
//printf("battery raw %g\n",v);
return v*ADC_factor/Divisor_battery;
}


void start_current(uint8_t sel) 
{
static uint8_t oldsel=0xff;
if(sel==oldsel) return;

GPIO_Set_Mode(GPIOA,5,GPIO_Mode_OUT);//ext sync signal
GPIO_Set_Mode(GPIOA,8,GPIO_Mode_IN);
GPIO_Set_Mode(GPIOA,11,GPIO_Mode_IN);
GPIO_Set_Mode(GPIOA,12,GPIO_Mode_IN);
GPIO_Set_Mode(GPIOA,15,GPIO_Mode_IN);
GPIO_SetBits(GPIOA, GPIO_Pin_8); //!goes to pnp
GPIO_SetBits(GPIOA, GPIO_Pin_15); //!goes to pnp
GPIO_ResetBits(GPIOA, GPIO_Pin_11); //goes directly
GPIO_ResetBits(GPIOA, GPIO_Pin_12); //goes directly
GPIO_ResetBits(GPIOA, GPIO_Pin_5); //goes directly

switch(sel)
	{ 
case 1: 
	GPIO_SetBits(GPIOA, GPIO_Pin_12);
	GPIO_Set_Mode(GPIOA,12,GPIO_Mode_OUT);
	break;
case 2:
	GPIO_SetBits(GPIOA, GPIO_Pin_11);
	GPIO_Set_Mode(GPIOA,11,GPIO_Mode_OUT);
	break;
case 3:
	GPIO_ResetBits(GPIOA, GPIO_Pin_8);
	GPIO_Set_Mode(GPIOA,8,GPIO_Mode_OUT);
	break;
case 4:
	GPIO_ResetBits(GPIOA, GPIO_Pin_15);
	GPIO_Set_Mode(GPIOA,15,GPIO_Mode_OUT);
	break;
case 5:
	GPIO_SetBits(GPIOA, GPIO_Pin_5);
	break;
case 0:
default:
	break;
	}
oldsel=sel;
}



void kbd_spi1_read(char *s, uint16_t len)
{
while(len--)
	{
	delay_ms(150);
	*s++ = kbd_spi1_getchar();
	*s=0;
	lcd_spi1_print(s-1);
	}
}



void init_current_switches(uint8_t slowedges)
{
RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOA, ENABLE);
GPIO_InitTypeDef GPIO_InitStructure;
GPIO_InitStructure.GPIO_Pin =  GPIO_Pin_5|GPIO_Pin_8|GPIO_Pin_15|GPIO_Pin_11|GPIO_Pin_12;
GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN;
GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
GPIO_InitStructure.GPIO_Speed = slowedges?GPIO_Speed_2MHz:GPIO_Speed_50MHz;
GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
GPIO_Init(GPIOA, &GPIO_InitStructure);
start_current(0);
}


void init_ad8253(void)
{
RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOC, ENABLE);
GPIO_InitTypeDef GPIO_InitStructure;
GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
GPIO_InitStructure.GPIO_Pin =  GPIO_Pin_14|GPIO_Pin_15;
GPIO_Init(GPIOC, &GPIO_InitStructure);
//no amplification
GPIO_ResetBits(GPIOC, GPIO_Pin_14);
GPIO_ResetBits(GPIOC, GPIO_Pin_15);
}

void init_capacitor(void)
{
//gpio for capacitor discharge
RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOC, ENABLE);
GPIO_InitTypeDef GPIO_InitStructure;
GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
GPIO_InitStructure.GPIO_Pin =  GPIO_Pin_13;
GPIO_Init(GPIOC, &GPIO_InitStructure);
//turn transistor off
GPIO_ResetBits(GPIOC, GPIO_Pin_13);

//timer5 to measure charging time
RCC_APB1PeriphResetCmd(RCC_APB1Periph_TIM5, ENABLE);
RCC_APB1PeriphResetCmd(RCC_APB1Periph_TIM5, DISABLE);
RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM5, ENABLE);
TIM_TimeBaseInitTypeDef TIM_TimeBaseInitStructure;
TIM_TimeBaseStructInit(&TIM_TimeBaseInitStructure);
TIM_TimeBaseInitStructure.TIM_Prescaler=1;
TIM_TimeBaseInit(TIM5,&TIM_TimeBaseInitStructure);
TIM_InternalClockConfig(TIM5);
TIM_Cmd(TIM5,ENABLE);
}

void gain_ad8253(uint8_t g)
{
GPIO_WriteBit(GPIOC,GPIO_Pin_14,g&1);
GPIO_WriteBit(GPIOC,GPIO_Pin_15,g&2);
}

void init_measurement(uint8_t cs, uint8_t gs)
{
RCC_APB1PeriphClockCmd(RCC_APB1Periph_PWR,ENABLE);
PWR_SDADCAnalogCmd(PWR_SDADCAnalog_1, ENABLE);
RCC_APB1PeriphClockCmd(RCC_APB1Periph_PWR, ENABLE);
RCC_APB2PeriphClockCmd(RCC_APB2Periph_SDADC1, ENABLE);
  /* NOTE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 * Select External reference: The reference voltage selection is available
     only in SDADC1 and therefore to select the VREF for SDADC2/SDADC3, SDADC1
     clock must be already enabled */

SDADC_VREFSelect(SDADC_VREF_VREFINT2);

uint8_t r;
r=init_SDADC1(current_sense_channel[cs],current_sense_gain[cs]);
if(r)  printf("error %d in init_SDADC1\n",r);
r=init_SDADC2(SDADC_Channel_6,sdadc_gain_list[gs]);
if(r)  printf("error %d in init_SDADC2\n",r);
delay_us(100);
}


jmp_buf env;
volatile uint8_t keypressed=0;

void kbd_callback(void)
{
keypressed=1;
}


//MEASUREMENT DEFAULTS
#define N_MEASUREMENT 4
typedef enum{menu=0,resistor=1,capacitor=2,inductor=3} MEASUREMENT;
MEASUREMENT measurement=resistor;
const char charlabel[N_MEASUREMENT+1]= "-RCL";
uint8_t currentselect=1;
uint8_t autorange=1;
uint8_t capacitance_preference=0;
uint16_t nrep1c=1;
uint16_t nrep1f=10;
uint8_t ad8253_gainselect=0;
uint8_t sdadc2_gainselect=0;
float vtarget=0.1;
#define nrep1c_init 100
#define nrep1c_max 10000
#define DIODE_THR 0.4


#define E_OK 0
//SDADC overflow
#define E_OVERFLOW 1
//low precision due to small SDADC readout
#define E_LOWPREC 2
//voltage drop on the DUT is too big, current flows through the protection diode
#define E_DIODE 4
#define E_OVERFLOWX 8
#define E_LOWPRECX 16


uint8_t measure_resistor(uint16_t nrep1, uint16_t nrep2, uint16_t delay, float *ix, float *ux)
{
int64_t xon0=0; int64_t xon1=0;
int64_t xoff0=0; int64_t xoff1=0;
uint16_t i;
uint8_t retstat=E_OK;
for(i=0; i<nrep2; ++i)
	{
	int32_t ron,roff;
	start_current(currentselect);  delay_ms(delay);
	xon0 += read_SDADC(SDADC1,nrep1);
	ron=read_SDADC(SDADC2,nrep1);
	xon1 += ron; 
	//delay_ms(3000);//just for calibration to read the multimeter
	if(ron >= 32766L*nrep1) {*ix= *ux=0.; return E_OVERFLOW;}

	start_current(0); delay_ms(delay);
	xoff0 += read_SDADC(SDADC1,nrep1);
	roff=read_SDADC(SDADC2,nrep1);
	xoff1 += roff; 
	//delay_ms(3000);//just for calibration to read the multimeter

	if(ron-roff <= 8192L*nrep1) retstat = E_LOWPREC;

	//poll if there was an interrupt, we cannot do longjmp directly from the interrupt handler!
	if(keypressed) longjmp(env,1);
	}

int64_t xdiff0;
int64_t xdiff1;
xdiff0 = xon0-xoff0;
xdiff1 = xon1-xoff1;
float current;
current = xdiff0;
current *= VREFINT2/(nrep1*nrep2*65536.);
current /= current_sense_resistor[currentselect]*current_sense_factor[currentselect];
printf("%%range=%f NREP %d %d ON  %lld %lld OFF %lld %lld DIFF %lld %lld\n",100.*xon1/nrep1/nrep2/32767.,nrep1,nrep2,xon0,xon1,xoff0,xoff1,xdiff0,xdiff1);

float voltage;
voltage = xdiff1;
voltage *= VREFINT2/(nrep1*nrep2*65536.);
voltage /= Divisor_direct*sdadc_gain_factors[sdadc2_gainselect]*ad8253_gain_factors[ad8253_gainselect];
if(voltage>DIODE_THR) retstat = E_DIODE;

*ix=current;
*ux=voltage;
return retstat;
}


uint8_t measure_capacitor(uint16_t nrep1, uint16_t nrep1fast, uint16_t nrep2, float *rx, float *cx, float *vx,float *tx, float*ix)
{
int64_t i1=0;
int64_t i2=0;
int64_t i3=0;
int64_t v1=0;
int64_t v2=0;
int64_t v3=0;
int32_t v1fast=0;
int32_t v2fast=0;
int32_t v3fast=0;
uint16_t i;
uint8_t retstat=E_OK;

float timer=0;
float timerfast=0;
float vinit;


init_ADC(ADC_Channel_1);

for(i=0; i<nrep2; ++i)
	{
	uint8_t i;

	//long discharge after current and capacitance measurement
	start_current(0); delay_ms(5);
	GPIO_SetBits(GPIOC, GPIO_Pin_13);
	i=0;
	delay_ms(30);
	do {
		if(keypressed) longjmp(env,1);
		++i; delay_ms(10); 
		vinit = read_SDADC(SDADC2,1);
		vinit *= VREFINT2/65536.;
		vinit /= Divisor_direct*sdadc_gain_factors[sdadc2_gainselect]*ad8253_gain_factors[ad8253_gainselect];
		}
	while(i<40 && vinit>0.005);
	GPIO_ResetBits(GPIOC, GPIO_Pin_13);
	delay_us(10);

	v1fast +=read_ADC(nrep1fast*10);
	start_current(currentselect); 
	TIM5->CNT=0;
	delay_us(1);
	uint32_t readoutfast = read_ADC(nrep1fast);
	start_current(0); 
	timerfast += TIM5->CNT;
	if(readoutfast>= 0xffe*nrep1fast) {*rx=  0./0.; return retstat|E_OVERFLOWX;}
	v2fast += readoutfast;
	delay_us(10);
	v3fast += read_ADC(nrep1fast*10);

	//short discharge after esr measurement
	start_current(0); delay_us(10);
        GPIO_SetBits(GPIOC, GPIO_Pin_13);
        delay_ms(10);
        GPIO_ResetBits(GPIOC, GPIO_Pin_13);
	delay_us(10);

	int32_t readout0 = read_SDADC(SDADC2,nrep1);
	v1+= readout0; 
	i1+= read_SDADC(SDADC1,nrep1);

	//measure just the current separately, as well as capacity
	start_current(currentselect);
	TIM5->CNT=0;
        i2+= read_SDADC(SDADC1,nrep1);
	int32_t readout=read_SDADC(SDADC2,nrep1);
        if(readout >= 32766L*nrep1) retstat|=E_OVERFLOW;
	v2 += readout;
	start_current(0);
	timer += TIM5->CNT;
	readout=read_SDADC(SDADC2,nrep1);
	if(readout >= 32766L*nrep1) retstat|=E_OVERFLOW;
	v3+= readout; 
	i3+= read_SDADC(SDADC1,nrep1);

	//poll if there was an interrupt, we cannot do longjmp directly from the interrupt handler!
	if(keypressed) longjmp(env,1);
	}

timer *= 2./((float)nrep2*SYSTEM_CLOCK);
timerfast *= 2./((float)nrep2*SYSTEM_CLOCK);

float current;
current = i2-(i1+i3)/2;
current *= VREFINT2/(nrep1*nrep2*65536.);
current /= current_sense_resistor[currentselect]*current_sense_factor[currentselect];

float vcharged;
float vfinal;
float vinitfast,vfinalfast;
float vdiff,vdifffast;
float range,rangefast;

vinitfast = v1fast;
vinitfast *= ADC_factor*VDD_analog/(nrep1fast*10*nrep2*4096.);
vinitfast /= Divisor_direct*ad8253_gain_factors[ad8253_gainselect];

vfinalfast = v3fast;
vfinalfast *= ADC_factor*VDD_analog/(nrep1fast*10*nrep2*4096.);
vfinalfast /= Divisor_direct*ad8253_gain_factors[ad8253_gainselect];


//printf("test %d %d %d\n",v1fast/nrep1fast/nrep2/10,v2fast/nrep1fast/nrep2,v3fast/nrep1fast/nrep2/10);
vdifffast = v2fast*10-(v1fast+v3fast)*.5;
vdifffast *= ADC_factor*VDD_analog/(nrep1fast*10*nrep2*4096.);
vdifffast /= Divisor_direct*ad8253_gain_factors[ad8253_gainselect];
rangefast = 100.*v2fast/nrep1fast/nrep2/4095.;
if(rangefast<= 9.5) retstat |= E_LOWPRECX;
if(rangefast<-10.) {*cx=*rx=0./0.; return E_OVERFLOWX;}

vfinal = v3;
vfinal *= VREFINT2/(nrep1*nrep2*65536.);
vfinal /= Divisor_direct*sdadc_gain_factors[sdadc2_gainselect]*ad8253_gain_factors[ad8253_gainselect];

float vcharging = v2;
vcharging *= VREFINT2/(nrep1*nrep2*65536.);
vcharging /= Divisor_direct*sdadc_gain_factors[sdadc2_gainselect]*ad8253_gain_factors[ad8253_gainselect];

vinit=v1;
vinit *= VREFINT2/(nrep1*nrep2*65536.);
vinit /= Divisor_direct*sdadc_gain_factors[sdadc2_gainselect]*ad8253_gain_factors[ad8253_gainselect];

vcharged = vfinal-vinit;
vdiff = vcharging-0.5*(vinit+vfinal);
if(vfinal > DIODE_THR|| vfinalfast > DIODE_THR) retstat |= E_DIODE;

range = 100.*v3/nrep1/nrep2/32767.;
if(range<33.) retstat |= E_LOWPREC;
if(range<-10.) retstat |= E_OVERFLOW;

printf("%%ranges=%.2f %.2f, nrep2=%d nrep1c=%d, times=%g %g I= %g Vinitfast=%g Vdifffast=%g Vfinalfast=%g Vinit=%g Vfinal=%g Vcharged=%g Vdiffslow=%g\n",rangefast,range,nrep2,nrep1,timerfast,timer,current,vinitfast,vdifffast,vfinalfast,vinit,vfinal,vcharged,vdiff);


*vx = vcharged;
*ix = current;
*tx=timer;
*rx= vdifffast/current;
if(retstat&E_OVERFLOWX) *rx = 0./0.; 
*cx= current*timer/vcharged;
if(*cx<0) *cx=1./0.;
if(retstat&E_DIODE) *cx = 0./0.;
if(retstat&E_OVERFLOW) *cx = 0./0.; 
return retstat;
}



void set_currentselect(uint8_t cs)
{
if(cs>NCURRENT) return;
static uint8_t oldcurrentselect=0;
currentselect=cs;
if(currentselect!=oldcurrentselect)
	{
	start_current(0); delay_ms(200);
	uint8_t r;
	r=init_SDADC1(current_sense_channel[currentselect],current_sense_gain[currentselect]);
	if(r) printf("error in init_SDADC1\n");
	oldcurrentselect=currentselect;
	}
printf("Currentselect is now %d\n",currentselect);
}


void set_voltage_gains(uint8_t g1,uint8_t g2)
{
if(g1>3||g2>4) return;
ad8253_gainselect=g1;
sdadc2_gainselect=g2;
gain_ad8253(ad8253_gainselect);
start_current(0); delay_ms(200);
uint8_t r;
r= init_SDADC2(SDADC_Channel_6,sdadc_gain_list[sdadc2_gainselect]);
if(r) printf("error in init_SDADC2\n");
r=0;
printf("Voltage gains are now %d %d\n",ad8253_gainselect,sdadc2_gainselect);
}



uint8_t increase_gain(void)
{
uint8_t err=0;
if(sdadc2_gainselect<4) ++sdadc2_gainselect;
else
	{
	if(ad8253_gainselect<3) 
		{
		sdadc2_gainselect=0;
		++ad8253_gainselect;
		}
	else err=1;
	}
if(!err) set_voltage_gains(ad8253_gainselect,sdadc2_gainselect);
return err;
}

uint8_t decrease_gain(void)
{
uint8_t err=0;
if(sdadc2_gainselect>0) --sdadc2_gainselect;
else
	{
	if(ad8253_gainselect>0)
		{
		sdadc2_gainselect=4;
		--ad8253_gainselect;
		}
	else err=1;
	}
if(!err)set_voltage_gains(ad8253_gainselect,sdadc2_gainselect);
return err;
}



int main(void)
{
MEASUREMENT oldmeasurement=0;

usart1_init(115200);
printsimple(USART1, "Milliohmmeter reset\n"); 
MCP23S17_reset();
delay_ms(30);
lcd_spi1_init();
delay_ms(50);
kbd_spi1_init();
delay_ms(20);
lcd_spi1_clear();
lcd_spi1_print("Milliohmmeter");
float vbat;
{
char c[17];
sprintf(c,"Vbat = %.2fV",vbat=vbattery());
printf("%s\n",c);
lcd_spi1_cursor(1,0);
lcd_spi1_print(c);
delay_ms(2000);
}


init_current_switches(0);
init_ad8253();
init_capacitor();

kbd_trap_line(3,kbd_callback);


gain_ad8253(ad8253_gainselect);
init_measurement(currentselect,sdadc2_gainselect);


while (1)
	{
	uint8_t fromwhere = setjmp(env);
	if(fromwhere) print(USART1,"after longjmp\n");

	if(inuart1line)
		{
		switch(inuart1[0])
			{
			case 't': //number of adc reads while charging capacitor
				{
				int tmp;
				if(1==sscanf(inuart1+1,"%d",&tmp)) nrep1f=tmp;
				}
				break;
			case 'A':
			case 'a': //autorange
				capacitance_preference= (inuart1[0]=='a'?0:1);
				autorange=1;  set_voltage_gains(0,0); nrep1c=capacitance_preference?nrep1c_init:1;
				break;
			case 'r':
				measurement = resistor;
				break;
			case 'l':
				measurement = inductor;
				break;
			case 'c':
				measurement = capacitor;
				break;
			case 'v': //test voltage for capacitance
				if(1!=sscanf(inuart1+1,"%f",&vtarget)) vtarget=100.;
				vtarget *= .001;
				if(vtarget<0. || vtarget>DIODE_THR) vtarget=0.2;
				break;
			case 'i':
			case '#': //test current 
				set_currentselect(inuart1[1]-'0');
                                if(autorange)
                                        {
                                        autorange=1;
					set_voltage_gains(0,0);
                                        }
				break;
			case '*': //select gains
				if(inuart1[1]=='*'||inuart1[1]=='#') {capacitance_preference=(inuart1[1]=='*'?0:1);autorange=1;  set_voltage_gains(0,0); nrep1c=capacitance_preference?nrep1c_init:1;}
				else {autorange=0; set_voltage_gains(inuart1[1]-'0',inuart1[2]-'0');}
				break;
			default:
				printf("Unknown command: %s\n",inuart1);
				break;
			}
		inuart1line=inuart1len=0;
		}

	if(keypressed) 
		{
		char key=decode_char(3,MCP23S17_Readinterrupt(1));
		switch(key)
			{
			case '#': //test current or voltage for capacitance
				{
				lcd_spi1_clear();
				lcd_spi1_print("#");
				char z[5];
				delay_ms(150);
				z[0] = kbd_spi1_getchar();
				z[1]=0;
				lcd_spi1_print(z);
				if(z[0]=='#') //setting vtarget
					{
					kbd_spi1_read(z,3); 
					if(1!=sscanf(z,"%f",&vtarget)) vtarget=100.;
                                	vtarget *= .001;
					if(vtarget<0. || vtarget>DIODE_THR) vtarget=0.2;
					break;
					}
				set_currentselect(z[0]-'0');
				if(autorange) 
					{
					autorange=1;
					nrep1c=capacitance_preference?nrep1c_init:1;
					set_voltage_gains(0,0);
					}
				}
				break;
			case '0': 
				{
				measurement = (measurement+1)%N_MEASUREMENT;
				}
				break;
			case '*': //select gains
				{
				lcd_spi1_clear();
                                lcd_spi1_print("*");
				char z[3];
				delay_ms(150);
				z[0] = kbd_spi1_getchar();delay_ms(200);
				z[1]=0;
				lcd_spi1_print(z);
				if(z[0]=='*'||z[0]=='#') {capacitance_preference=(z[0]=='*'?0:1);autorange=1; set_voltage_gains(0,0);nrep1c=capacitance_preference?nrep1c_init:1; break;}
				z[1] = kbd_spi1_getchar();delay_ms(50);
                                z[2]=0;
				lcd_spi1_print(z+1);
				autorange=0; set_voltage_gains(z[0]-'0',z[1]-'0');
				}
				break;
			default: break;
			}
		delay_ms(200);
		keypressed=0;
		kbd_trap_line(3,kbd_callback);
		}

	uint16_t nrep1,nrep2;
	
	if(measurement!=oldmeasurement) init_current_switches(measurement==inductor?1:0);

	switch(measurement)
		{
		case capacitor:
			{
			if(autorange==1) nrep2=2;
                        else    {nrep2=100000L/nrep1c; if(nrep2>100) nrep2=100;}
			char c[32];
			float rx,cx,vx,tx,ix;
			uint8_t err=measure_capacitor(nrep1c,nrep1f,nrep2,&rx,&cx,&vx,&tx,&ix); 
			vbat=vbattery();
			if(capacitance_preference==0)
			   {
			   if(err&E_OVERFLOWX)
				{
				if(!autorange) delay_ms(1000);
				else {autorange=1; if(ad8253_gainselect>0) --ad8253_gainselect;}
				set_voltage_gains(ad8253_gainselect,sdadc2_gainselect);
				goto endswitch;
				}
			    if(err& E_LOWPRECX)
				{
				if(autorange)
					{autorange=1; if(ad8253_gainselect<3) ++ad8253_gainselect;}
				set_voltage_gains(ad8253_gainselect,sdadc2_gainselect);
				goto endswitch;
				}
			    }
			if(err& E_OVERFLOW)
				{
				if(!autorange) delay_ms(1000);
				else 
					{
					autorange= capacitance_preference?1:2; 
					if(capacitance_preference) {if(decrease_gain()) autorange=2;}
					else
						{ 
						if(sdadc2_gainselect>0) --sdadc2_gainselect; 
						else if(nrep1c>1) --nrep1c; //shorten charging time
						}
					}
				set_voltage_gains(ad8253_gainselect,sdadc2_gainselect);
				goto endswitch;
				}
			if(capacitance_preference==1 && (vx < 0.9 * vtarget || vx > 1.1*vtarget)) {nrep1c = vtarget/vx*nrep1c; if(nrep1c>nrep1c_max) nrep1c=nrep1c_max;}
			if(err& E_LOWPREC)	
				{
				if(autorange) 
					{
					autorange=capacitance_preference?1:2; 
					if(capacitance_preference) {if(increase_gain()) autorange=2;}
                                        else 
						{
						if(sdadc2_gainselect<4) ++sdadc2_gainselect;
						else ++nrep1c;
						}
					}
				set_voltage_gains(ad8253_gainselect,sdadc2_gainselect);
				goto endswitch;
				}
			if(autorange) autorange=2;
			endswitch:
			if(capacitance_preference==0) sprintf(c,"R%c%.5g",(err&E_OVERFLOWX||err&E_LOWPRECX||err&E_DIODE?'!':'='),rx);
			else sprintf(c,"C%c%.5guF",(err&E_OVERFLOW||err&E_LOWPREC||err&E_DIODE?'!':'='),1e6*cx);
			printf("%s\n",c);
			if(oldmeasurement!=measurement) lcd_spi1_clear();
			lcd_spi1_cursor(0,0);
			lcd_spi1_printn(c,11);
			sprintf(c,"%c%1d%c%1d%1d",charlabel[measurement],currentselect,(autorange?(capacitance_preference?'a':'A'):'M'),ad8253_gainselect,sdadc2_gainselect);
			lcd_spi1_cursor(0,11);
			lcd_spi1_print(c);
			printf("%s\n",c);
			if(capacitance_preference==0) sprintf(c,"C%c%.0fuF",(err&E_OVERFLOW||err&E_LOWPREC||err&E_DIODE?'!':(capacitance_preference?'=':'!')),1e6*cx);
			else sprintf(c,"%.3g*%.3g/%.3g",tx*1000,ix*1000,vx*1000);
			printf("%s\n",c);
			lcd_spi1_cursor(1,0);
                        lcd_spi1_printn(c,16);
			}
			break;
		case inductor: //minimize ac frequency component, slower edges, otherwise as resistor
			if(autorange==1) {nrep1=500; nrep2=1;}
			else  {nrep1=2000; nrep2=2;}
			goto do_resistor;
			break;
		case resistor:
			if(autorange==1) {nrep1=100; nrep2=1;}
			else	{nrep1=100; nrep2=50;}
			do_resistor:
			{
			char c[32];
			float ix,ux;
			uint8_t err=measure_resistor(nrep1,nrep2,(measurement==inductor?200:10),&ix,&ux); 
			vbat=vbattery();
			float ohm;
			switch(err)
				{
				case E_OVERFLOW:
					if(!autorange) delay_ms(1000);
					else {autorange=decrease_gain()?2:1;}
					ohm=0./0.;
					break;
				case E_DIODE:
					if(autorange) autorange=2;
					ohm=1./0.;
					break;
				case E_LOWPREC:
					if(autorange) {autorange=increase_gain()?2:1;}
					ohm=ux/ix;
					break;
				default:
					if(autorange) autorange=2;
					ohm=ux/ix;
					break;
				}
			sprintf(c,"R%c%.5g",(err?'!':'='),ohm);
			printf("%s\n",c);
			if(oldmeasurement!=measurement) lcd_spi1_clear();
			lcd_spi1_cursor(0,0);
			lcd_spi1_printn(c,11);
			sprintf(c,"%c%1d%c%1d%1d",charlabel[measurement],currentselect,(autorange?(capacitance_preference?'a':'A'):'M'),ad8253_gainselect,sdadc2_gainselect);
			lcd_spi1_cursor(0,11);
			lcd_spi1_print(c);
			printf("%s\n",c);
			sprintf(c,"%.5g/%.5g",ux*1000,ix*1000);
			printf("%s\n",c);
			lcd_spi1_cursor(1,0);
                        lcd_spi1_printn(c,16);
			}
			break;
		case menu:
		default:
			if(oldmeasurement!=measurement) lcd_spi1_clear();
			lcd_spi1_cursor(0,0);
			lcd_spi1_print("MENU (0=back)");
			delay_ms(150);
			char z=kbd_spi1_getchar();
			switch(z)
				{
				case '9':
					spi1_light ^= 1<<7;
					break;
				case '0':
				default:
					measurement=resistor;
					break;
				}
			delay_ms(150);
			keypressed=0;
			kbd_trap_line(3,kbd_callback);
			break;
		}
	oldmeasurement=measurement;
	printf("battery=%fV\n",vbat);
	if(vbat<LOWBAT) 
		{
		start_current(0);
		spi1_light=0;
		lcd_spi1_clear();
		lcd_spi1_print("Battery low!");
		lcd_spi1_cursor(1,0);
		char c[16];
		sprintf(c,"V=%.2fV",vbat);
		lcd_spi1_print(c);
		PWR_EnterSTANDBYMode();
		}
	}//while
}

