/*
    Copyright (C) 2020 Jiri Pittner <jiri@pittnerovi.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

//simple analysis of NMEA messages, library and standalone main()


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <ctype.h>

#include "nmea.h"

int nmea_checksum_error(char *msg)
{
if(msg[0]!='$') return 1;
char *pstar=strchr(msg,'*');
if(!pstar) return 2;
char *p;
int chk=0;
for(p=msg+1; p<pstar; ++p) chk ^= *p;
int sum;
if(1!=sscanf(pstar+1,"%x",&sum)) return 3;
if(chk!=sum) return 4;
return 0;
}

//return 0 on OK, nonzero on errors
//update GPS data if new ones arrived in the current sentence
int nmea_process(char *msg, char *msg_type, time_t *last_fix,time_t *gps_time, float *gps_position, float *ground_speed, float *heading, int *nsat, float *accuracy)
{
//checksum error
if(nmea_checksum_error(msg)) return 1;
char *sentence_type=strsep(&msg,",");
if(!sentence_type) return 2; //no token found
if(msg_type)
	{
	if(strncmp(sentence_type,msg_type,3)) return 3; //not a GPS message
	}

if(!strncmp(sentence_type+3,"GSV",3)) return 0; //satelites in view
if(!strncmp(sentence_type+3,"VTG",3)) return 0; //contained in RMC anyway
if(!strncmp(sentence_type+3,"GLL",3)) return 0; //contained in RMC anyway
if(!strncmp(sentence_type+3,"GSA",3)) return 0; //more detailed gps accuracy - not processed presently


if(!strncmp(sentence_type+3,"RMC",3)) //position and time
	{
	int has_fix=1;
	char *nexttoken;
	struct tm zulutime;
	nexttoken=strsep(&msg,","); //time
	if(nexttoken==NULL) return 1;
	int r;
	r=sscanf(nexttoken,"%2d%2d%2d",&zulutime.tm_hour,&zulutime.tm_min,&zulutime.tm_sec);
	if(r!=3) return 0;
	nexttoken=strsep(&msg,","); //status
	if(nexttoken==NULL) return 1;
	if(toupper(*nexttoken)=='V') return 0; 
	nexttoken=strsep(&msg,","); //latitude
	if(nexttoken==NULL) return 1;
	int degs;
	float mins;
	r=sscanf(nexttoken,"%2d%f",&degs,&mins);
	if(r!=2) has_fix=0;
	else gps_position[0]=degs+mins/60.f;
	nexttoken= strsep(&msg,",");
	if(nexttoken==NULL) return 1;
        if(has_fix && toupper(*nexttoken)=='S') gps_position[0]= -gps_position[0];
	nexttoken=strsep(&msg,","); //longitude
	if(nexttoken==NULL) return 1;
        r=sscanf(nexttoken,"%3d%f",&degs,&mins);
        if(r!=2) has_fix=0;
        else gps_position[1]=degs+mins/60.f;
        nexttoken= strsep(&msg,",");
	if(nexttoken==NULL) return 1;
        if(has_fix && toupper(*nexttoken)=='W') gps_position[1]= -gps_position[1];
	nexttoken=strsep(&msg,","); //ground speed
	if(nexttoken==NULL) return 1;
	if(has_fix) 
		{
		float gs;
		if(1==sscanf(nexttoken,"%f",&gs)) *ground_speed = gs *1852.f/3600.f; //from knots to m/s
		}
	nexttoken=strsep(&msg,","); //heading
	if(nexttoken==NULL) return 1;
	float h;
	r=sscanf(nexttoken,"%f",&h);
	if(r==1 && h<360.f) *heading=h;
	nexttoken=strsep(&msg,","); //date
	if(nexttoken==NULL) return 1;
	r=sscanf(nexttoken,"%2d%2d%2d",&zulutime.tm_mday,&zulutime.tm_mon,&zulutime.tm_year);
        if(r!=3) return 0;
	zulutime.tm_mon--; //counted from 0
	zulutime.tm_year += 100; //counted from 1900	
	zulutime.tm_isdst=0;
	tzname[0] = tzname[1] = "GMT";
	setenv("TZ", "UTC", 1);
	#ifdef _XOPEN_SOURCE
	timezone = 0;
	daylight = 0;
	#endif
	*gps_time = mktime(&zulutime);
	if(has_fix)  *last_fix = *gps_time;
	}

if(!strncmp(sentence_type+3,"GGA",3)) //altitude and accuracy
	{
        char *nexttoken;
        //struct tm zulutime;
        nexttoken=strsep(&msg,","); //time
	if(nexttoken==NULL) return 1;
	int r;
        //int centisec;
        //r=sscanf(nexttoken,"%2d%2d%2d.%2d",&zulutime.tm_hour,&zulutime.tm_min,&zulutime.tm_sec,&centisec); if(r!=4) return 0;
	nexttoken=strsep(&msg,","); //latitude
	if(nexttoken==NULL) return 1;
        int degs;
        float mins;
        r=sscanf(nexttoken,"%2d%f",&degs,&mins);
        if(r!=2) return 0;
        nexttoken= strsep(&msg,",");
	if(nexttoken==NULL) return 1;
        nexttoken=strsep(&msg,","); //longitude
	if(nexttoken==NULL) return 1;
        r=sscanf(nexttoken,"%3d%f",&degs,&mins);
        if(r!=2) return 0;
        nexttoken=strsep(&msg,","); 
	if(nexttoken==NULL) return 1;
        nexttoken=strsep(&msg,","); //fix
	if(nexttoken==NULL) return 1;
	if(*nexttoken=='0') return 0;
	nexttoken=strsep(&msg,","); //nsat
	if(nexttoken==NULL) return 1;
	sscanf(nexttoken,"%d",nsat);
	nexttoken=strsep(&msg,","); //accuracy
	if(nexttoken==NULL) return 1;
	sscanf(nexttoken,"%f",accuracy);
	nexttoken=strsep(&msg,","); //altitude
	if(nexttoken==NULL) return 1;
	float alt;
	if(1!=sscanf(nexttoken,"%f",&alt)) return 0;
	nexttoken=strsep(&msg,","); //altitude units
	if(nexttoken==NULL) return 1;
	if(toupper(*nexttoken)!='M') return 0;
	gps_position[2]=alt;
	}



return 0;
}



#ifdef  NMEA_STANDALONE
void main() //read nmea from standard input and process it
{
char line[1024];
time_t last_fix=0;

float gps_position[3];
memset(gps_position,0,3*sizeof(float));
time_t gps_time=0;;
float ground_speed=0.;
float heading=0;
int nsat=0;
float accuracy=0;
while(fgets(line,1024,stdin))
	{
	int r=nmea_process(line,NULL,&last_fix,&gps_time,gps_position,&ground_speed,&heading,&nsat,&accuracy);
	struct tm *tmp;
	tmp = gmtime(&gps_time);
        char times[64];
        strftime(times,64,"%c %Z",tmp);
	printf("return %d date %s last %ld date %ld position %f %f altitude %f speed %f hdg %f nsat %d accuracy %f\n",r,times,last_fix,gps_time,gps_position[0],gps_position[1],gps_position[2],ground_speed,heading,nsat,accuracy);
	}
}
#endif
