/*
    Hardware-dependent sensor routines for avionics

    Copyright (C) 2020 Jiri Pittner <jiri@pittnerovi.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "delay.h"
#include "pressure_sensors.h"
#include "avi_calibration.h"

/*HW setup:
PE15=SDADC2_0= VCC5A reference
PE14=SDADC2_1= Static  MPX4115A
PE13=SDADC2_2= Pitot  MPX5010DP
PE12=SDADC2_3= AOA Pitot  MPX5010DP
PE11=SDADC2_4= not used
all with divisors 330/(220+330)

BMP180 at I2C1 - TODO @@@@@@@@@
*/


#define SDADC_INIT_TIMEOUT   2000
#define SDADC_CAL_TIMEOUT    (40*30720L) /*  ~5.12 ms at 6 MHz  in a single calibration sequence */
//this is memory intensive, better keep SDADC2_NREP low and do extra lowpassing
#define SDADC2_alpha 0.99f
#define SDADC2_NREP 32
#define SDADC2_vref 3.30f
#define SDADC2_R_divisor (550.f/330.f)


const uint32_t SDADC2_gains[3] = {SDADC_Gain_1,SDADC_Gain_1,SDADC_Gain_1}; //there are just 3 configurations available
const float SDADC2_gainfactors[SDADC2_NCHAN]={1.f,1.f,1.f,1.f};
const uint32_t SDADC2_configs[SDADC2_NCHAN] = {0,0,0,0}; //use config0 for all
const uint32_t SDADC2_channels[SDADC2_NCHAN] = {SDADC_Channel_0,SDADC_Channel_1,SDADC_Channel_2,SDADC_Channel_3};

volatile float vsdadc2[SDADC2_NCHAN]={0,0,0,0};

volatile uint32_t SDADC2_data[SDADC2_NREP*SDADC2_NCHAN]; //DMA buffer
volatile static uint8_t novice=1;

//interrupt to process data after complete DMA transfer
void DMA2_Channel4_IRQHandler(void)
{
DMA_ClearITPendingBit(DMA2_IT_TC4);
static volatile uint8_t SDADC2_irqpass=0;
++SDADC2_irqpass;
uint16_t i;
uint32_t sum[SDADC2_NCHAN];
uint16_t hist[SDADC2_NCHAN];
memset(sum,0,SDADC2_NCHAN*sizeof(uint32_t));
memset(hist,0,SDADC2_NCHAN*sizeof(uint16_t));
for(i=0; i<SDADC2_NREP*SDADC2_NCHAN; ++i)
	{
	uint8_t ch = (SDADC2_data[i]>>24); //no need to remap the channel index as we are using channels 0,1,2,3
	if(ch>=SDADC2_NCHAN) continue; //this should not happen
	uint16_t tmp = (uint16_t)32768 + (uint16_t)(SDADC2_data[i]&0xffff);
		{
		++hist[ch];
		sum[ch] += tmp;
		}
	}


for(i=0; i<SDADC2_NCHAN; ++i)
	{
	float newvalue = sum[i]*SDADC2_vref/hist[i]/SDADC2_gainfactors[i]/65536.;
	if(novice)
		{
		vsdadc2[i] = newvalue;
		}
	else
		{
		vsdadc2[i] *= SDADC2_alpha;
		vsdadc2[i] += (1.f-SDADC2_alpha)*newvalue;
		}
	}
novice=0;

/*debug
#define SDADC2_printmod 0xff
if((SDADC2_irqpass&SDADC2_printmod)==SDADC2_printmod) 
    {
    for(i=0; i<SDADC2_NCHAN; ++i)
        {
        printf("sdadc2_ch%d v= %5.3f ",i,vsdadc2[i]);
        }
    printf("\n");
    }
*/
}



uint8_t init_SDADC2(uint8_t use_dma, const uint32_t *gains, const uint32_t *configs, const uint32_t *channels, uint32_t *buffer, uint16_t blocksize, uint8_t nchan)
{
uint32_t SDADCTimeout = 0;

SDADC_DeInit(SDADC2); 

 RCC_APB1PeriphClockCmd(RCC_APB1Periph_PWR,ENABLE);

//reference voltage common to all SDADCs
//SDADC_VREFSelect(SDADC_VREF_VDDA); // SDADC_VREF_Ext requires populating R112, with SDADC_VREF_VREFINTx  R112 must NOT be populated
  /* NOTE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 * Select External reference: The reference voltage selection is available
     only in SDADC1 and therefore to select the VREF for SDADC2/SDADC3, SDADC1
     clock must be already enabled */
RCC_APB2PeriphClockCmd(RCC_APB2Periph_SDADC1, ENABLE);
SDADC_VREFSelect(SDADC_VREF_VDDA); // SDADC_VREF_Ext requires populating R112, with SDADC_VREF_VREFINTx  R112 must NOT be populated


  PWR_SDADCAnalogCmd(PWR_SDADCAnalog_2, ENABLE);
  RCC_APB2PeriphClockCmd(RCC_APB2Periph_SDADC2, ENABLE);
  RCC_SDADCCLKConfig(RCC_SDADCCLK_SYSCLK_Div12); /* Set the SDADC divider: The SDADC should run max@6MHz, we run at 6.144 */
  RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOE, ENABLE);

  /* allowed SDADC2 channels 0,1,2,3 single ended on pe15,14,13,12 respectively */
  GPIO_InitTypeDef GPIO_InitStructure;
  GPIO_InitStructure.GPIO_Mode  = GPIO_Mode_AN;
  GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
  GPIO_InitStructure.GPIO_Pin = GPIO_Pin_12|GPIO_Pin_13|GPIO_Pin_14|GPIO_Pin_15;
  GPIO_Init(GPIOE, &GPIO_InitStructure);

  SDADC_Cmd(SDADC2, ENABLE);
  delay_ms(5);

  /* Enter initialization mode */
  SDADC_InitModeCmd(SDADC2, ENABLE);
  SDADCTimeout = SDADC_INIT_TIMEOUT;
  /* wait for INITRDY flag to be set */
  while((SDADC_GetFlagStatus(SDADC2, SDADC_FLAG_INITRDY) == RESET) && (--SDADCTimeout != 0)) ;

  if(SDADCTimeout == 0)
  {
    /* INITRDY flag can not set */
    return 1;
  }


//channel to pin mapping is in STM32F373 datasheet (not in reference)
SDADC_InitTypeDef SDADC_InitStruct;
SDADC_InitStruct.SDADC_Channel = SDADC_Channel_1; //here only one regular channel can be specified
SDADC_InitStruct.SDADC_ContinuousConvMode = use_dma?ENABLE:DISABLE;
SDADC_InitStruct.SDADC_FastConversionMode = DISABLE;
SDADC_Init(SDADC2, &SDADC_InitStruct);


SDADC_AINStructTypeDef SDADC_AINStructure;
SDADC_AINStructInit(&SDADC_AINStructure);
SDADC_AINStructure.SDADC_InputMode = SDADC_InputMode_SEZeroReference;
SDADC_AINStructure.SDADC_Gain = gains[0];
SDADC_AINStructure.SDADC_CommonMode = SDADC_CommonMode_VSSA; 
SDADC_AINStructure.SDADC_Offset = 0;
SDADC_AINInit(SDADC2, SDADC_Conf_0, &SDADC_AINStructure);
SDADC_AINStructure.SDADC_Gain = gains[1];
SDADC_AINInit(SDADC2, SDADC_Conf_1, &SDADC_AINStructure);
SDADC_AINStructure.SDADC_Gain = gains[2];
SDADC_AINInit(SDADC2, SDADC_Conf_2, &SDADC_AINStructure);


uint8_t i;
for(i=0; i<nchan; ++i) SDADC_ChannelConfig(SDADC2, channels[i], configs[i]); 

SDADC_CalibrationSequenceConfig(SDADC2, SDADC_CalibrationSequence_3);  //needed since we might  use all three configs

SDADC_InitModeCmd(SDADC2, DISABLE);

uint32_t tmp=0;
for(i=0; i<nchan; ++i) tmp |= channels[i];
SDADC_InjectedChannelSelect(SDADC2,tmp);


SDADC_StartCalibration(SDADC2);
SDADCTimeout = SDADC_CAL_TIMEOUT*3;
while((SDADC_GetFlagStatus(SDADC2, SDADC_FLAG_EOCAL) == RESET) && (--SDADCTimeout != 0)) ;
if(SDADCTimeout == 0)
  {
    return 2;
  }

//check the offsets
{
uint32_t tmp = 0;
tmp = (uint32_t)((uint32_t)SDADC2 + 0x00000020);
uint16_t offset0 = *(__IO uint32_t *) (tmp);
tmp = (uint32_t)((uint32_t)SDADC2 + 0x00000024);
uint16_t offset1 = *(__IO uint32_t *) (tmp);
tmp = (uint32_t)((uint32_t)SDADC2 + 0x00000028);
uint16_t offset2 = *(__IO uint32_t *) (tmp);
printf("SDADC2 offsets %d %d %d\n",offset0,offset1,offset2);
}


if(!use_dma) return 0;


RCC_AHBPeriphClockCmd(RCC_AHBPeriph_DMA2, ENABLE);

static DMA_InitTypeDef DMA_InitStruct;

  DMA_StructInit(&DMA_InitStruct);
  DMA_InitStruct.DMA_PeripheralBaseAddr = (uint32_t)&(SDADC2->JDATAR);
  DMA_InitStruct.DMA_MemoryBaseAddr = (uint32_t)buffer;
  DMA_InitStruct.DMA_DIR = DMA_DIR_PeripheralSRC;
  DMA_InitStruct.DMA_BufferSize = blocksize*nchan;
  DMA_InitStruct.DMA_PeripheralInc = DMA_PeripheralInc_Disable;
  DMA_InitStruct.DMA_MemoryInc = DMA_MemoryInc_Enable;
  DMA_InitStruct.DMA_PeripheralDataSize = DMA_PeripheralDataSize_Word;
  DMA_InitStruct.DMA_MemoryDataSize = DMA_MemoryDataSize_Word;
  DMA_InitStruct.DMA_Mode = DMA_Mode_Circular;
  DMA_InitStruct.DMA_Priority = DMA_Priority_Medium;
  DMA_InitStruct.DMA_M2M = DMA_M2M_Disable;
  DMA_Init(DMA2_Channel4, &DMA_InitStruct);


DMA_Cmd(DMA2_Channel4, DISABLE);
DMA_ClearFlag(DMA2_FLAG_TC4); //clear transfer complete to allow next one
DMA_ClearFlag(DMA2_FLAG_TE4); //clear transfer error
DMA_Cmd(DMA2_Channel4, ENABLE);
DMA_ITConfig(DMA2_Channel4,DMA_IT_TC,ENABLE);


SDADC_DMAConfig(SDADC2,SDADC_DMATransfer_Injected,ENABLE);
SDADC_InjectedContinuousModeCmd(SDADC2,ENABLE);
SDADC_SoftwareStartInjectedConv(SDADC2);

NVIC_InitTypeDef NVIC_InitStructure;
NVIC_InitStructure.NVIC_IRQChannel = DMA2_Channel4_IRQn;
NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 0;
NVIC_InitStructure.NVIC_IRQChannelSubPriority = 0;
NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
NVIC_Init(&NVIC_InitStructure);

return 0;
}

uint32_t uread_SDADC(SDADC_TypeDef *SDADC, uint16_t nrep)
{
//discard first one
SDADC_SoftwareStartConv(SDADC);
while(SDADC_GetFlagStatus(SDADC,SDADC_FLAG_REOC)!=SET);
SDADC_GetConversionValue(SDADC);

uint32_t r=0;
do
        {
        SDADC_SoftwareStartConv(SDADC);
        while(SDADC_GetFlagStatus(SDADC,SDADC_FLAG_REOC)!=SET);
        r += (uint16_t) ((uint16_t)32768+SDADC_GetConversionValue(SDADC));
        }
while(--nrep);
return r;
}


uint8_t init_pressure(void)
{
uint8_t r;
r=init_SDADC2(1,SDADC2_gains,SDADC2_configs,SDADC2_channels,SDADC2_data,SDADC2_NREP,SDADC2_NCHAN);
printf("init_SDADC2 returned %d\n",r);
if(r) {printf("needs hw reset %d\n",r); while(1); }
return r;
}


void read_pressures(float *p_static,float *p1, float *p2)

{
//vsdadc2 are voltages at the MCU
//compute voltages at the pressure sensors and the VCC5A input voltage
float VCC5A = SDADC2_R_divisor * vsdadc2[0];
float Vstatic = SDADC2_R_divisor * vsdadc2[1];
float Vpitot =  SDADC2_R_divisor * vsdadc2[2];
float Vpitot2 =  SDADC2_R_divisor * vsdadc2[3];

//compute pressures from voltages
*p_static = (Vstatic/VCC5A + 0.095f)/.009f*1000.f;  //in pascals
*p1=(Vpitot/VCC5A -0.04f)/0.09f*1000.f;
*p2=(Vpitot2/VCC5A -0.04f)/0.09f*1000.f;
}

void process_pressures(float p_static,float p1, float p2, float qnh, float temp, float rel_humidity, float *altitude, float *ias, float *tas, float *aoa, float *density_altitude)
{
float	p_pitot = p1 - p_pitot_calibration_offset;
float	p_pitot2 = p2 - p_pitot2_calibration_offset;

//printf("static pressure = %.0fPa, pitot1 pressure = %.0fPa, pitot2 pressure = %.0fPa\n",p_static,p_pitot,p_pitot2);

//compute altitude using standard atmosphere model, 15C at sea level
//to avoid solution of non-linear equation, use previous altitutde in the lapse rate
const float mgR=0.03416319473631;
const float lapse_rate = 0.0065;
float tmp=logf(qnh/p_static);
*altitude = tmp/mgR*(273.15f+15.f-0.5f*lapse_rate* *altitude);

//compute IAS from pitot dynamic pressure 
const float rho0 = 1.225f;
if(p_pitot<0) p_pitot=0;
*ias = sqrtf(p_pitot*2/rho0);

//compute air density from pressure, temperature and humidity 
//for formulas cf. https://en.wikipedia.org/wiki/Density_of_air
float abstemp = 273.15f + temp;

const float R_gas = 8.31446f;
const float M_w = 0.018016f; //molar mass h2o 
const float M_dry =  0.0289654f; //molar mass dry air

float p_saturated = 6.102f * powf(10.f,7.5f*temp/(temp+237.8f));
float p_water = rel_humidity * p_saturated;
float p_dry = p_static - p_water;

float rho= (p_dry * M_dry + p_water * M_w)/R_gas/abstemp;

//compute TAS; 
//neglecting compression effects at the low speeds of recreational aircraft
//just using the true air density 
*tas = sqrtf(p_pitot*2/rho);

//compute density altitude cf. https://en.wikipedia.org/wiki/Density_altitude
const float g_normal = 9.80665f;
const float stdabstemp = 273.15f+15.f;

*density_altitude = stdabstemp/lapse_rate*(1.f-powf((p_static/101325.f)/(abstemp/(stdabstemp)),lapse_rate*R_gas/(g_normal*M_dry-lapse_rate*R_gas)));

//compute AOA from two pitots at a certain angle (not supported presently)
*aoa = 0;
}

