/*
    Copyright (C) 2020 Jiri Pittner <jiri@pittnerovi.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.



*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include "rtc.h"
#include "stm32f37x_rcc.h"
#include "stm32f37x_rtc.h"
#include "stm32f37x_exti.h"
#include "stm32f37x_pwr.h"

#define RTC_CLOCK_SOURCE_LSE


static RTC_InitTypeDef RTC_InitStructure;
static __IO uint32_t AsynchPrediv = 0, SynchPrediv = 0;

void RTC_config(void)
{
printf("in RTC_config (if it hangs check the HW oscillator)\n");
RCC_APB1PeriphClockCmd(RCC_APB1Periph_PWR, ENABLE);
PWR_BackupAccessCmd(ENABLE);
RCC_BackupResetCmd(ENABLE);
RCC_BackupResetCmd(DISABLE);

#if defined (RTC_CLOCK_SOURCE_LSI)  /* LSI used as RTC source clock */
/* The RTC Clock may varies due to LSI frequency dispersion */
  /* Enable the LSI OSC */
  RCC_LSICmd(ENABLE);

  /* Wait till LSI is ready */
  while(RCC_GetFlagStatus(RCC_FLAG_LSIRDY) == RESET)
  {
  }


  /* Select the RTC Clock Source */
  RCC_RTCCLKConfig(RCC_RTCCLKSource_LSI);

  SynchPrediv = 0x18F;
  AsynchPrediv = 0x63;

#elif defined (RTC_CLOCK_SOURCE_LSE) /* LSE used as RTC source clock */
  /* Enable the LSE OSC */
  RCC_LSEConfig(RCC_LSE_ON);

  /* Wait till LSE is ready */
  while(RCC_GetFlagStatus(RCC_FLAG_LSERDY) == RESET)
  {
  }


  /* Select the RTC Clock Source */
  RCC_RTCCLKConfig(RCC_RTCCLKSource_LSE);

  SynchPrediv = 0xFF;
  AsynchPrediv = 0x7F;

#else
  #error Please select the RTC Clock source 
#endif /* RTC_CLOCK_SOURCE_LSI */

  /* Enable the RTC Clock */
  RCC_RTCCLKCmd(ENABLE);


  /* Wait for RTC APB registers synchronisation */
  RTC_WaitForSynchro();
printf("rtc_config complete\n");
}


void rtc_init(void)
{
#if defined (RTC_CLOCK_SOURCE_LSI)
    RCC_LSICmd(ENABLE);
#endif

printf("RTC_ReadBackupRegister %x\n",RTC_ReadBackupRegister(RTC_BKP_DR0));

if (RTC_ReadBackupRegister(RTC_BKP_DR0) != 0x32F2)
  {
    RTC_config();

printf("again RTC_ReadBackupRegister %x\n",RTC_ReadBackupRegister(RTC_BKP_DR0));

    /* Configure the RTC data register and RTC prescaler */
    RTC_InitStructure.RTC_AsynchPrediv = AsynchPrediv;
    RTC_InitStructure.RTC_SynchPrediv = SynchPrediv;
    RTC_InitStructure.RTC_HourFormat = RTC_HourFormat_24;

    /* Check on RTC init */
    if (RTC_Init(&RTC_InitStructure) == ERROR)
    {
      printf("\n\r        /!\\***** RTC Prescaler Config failed ********/!\\ \n\r");
    }

  }
  else
  {
printf("NOT calling RTC_config\n");
    /* Check if the Power On Reset flag is set */
    if (RCC_GetFlagStatus(RCC_FLAG_PORRST) != RESET)
    {
      printf("\r\n Power On Reset occurred....\n\r");
    }
    /* Check if the Pin Reset flag is set */
    else if (RCC_GetFlagStatus(RCC_FLAG_PINRST) != RESET)
    {
      printf("\r\n External Reset occurred....\n\r");
    }

    printf("\n\r No need to configure RTC....\n\r");

    /* Enable the PWR clock */
    RCC_APB1PeriphClockCmd(RCC_APB1Periph_PWR, ENABLE);

    /* Allow access to RTC */
    PWR_BackupAccessCmd(ENABLE);

    /* Wait for RTC APB registers synchronisation */
    RTC_WaitForSynchro();

    /* Clear the RTC Alarm Flag */
    RTC_ClearFlag(RTC_FLAG_ALRAF);

    /* Clear the EXTI Line 17 Pending bit (Connected internally to RTC Alarm) */
    EXTI_ClearITPendingBit(EXTI_Line17);

  }

}

void rtc_get(RTC_TimeTypeDef *rtctime, RTC_DateTypeDef *rtcdate)
{
  RTC_TimeStructInit(rtctime);
  RTC_DateStructInit(rtcdate);
  RTC_GetTime(RTC_Format_BIN, rtctime);
  RTC_GetDate(RTC_Format_BIN, rtcdate);
}


void rtc_set(RTC_TimeTypeDef *rtctime, RTC_DateTypeDef *rtcdate)
{
RTC_SetTime(RTC_Format_BIN, rtctime);
RTC_SetDate(RTC_Format_BIN, rtcdate);
/* Indicator for the RTC configuration */
RTC_WriteBackupRegister(RTC_BKP_DR0, 0x32F2);
}


int _gettimeofday(struct timeval *tv, struct timezone *tz)
{
if(!tv) return 0;
RTC_TimeTypeDef rtctime;
RTC_DateTypeDef rtcdate;
rtc_get(&rtctime,&rtcdate);
//printf("in _gettimeofday %d %d %d %d %d %d\n",rtcdate.RTC_Year,rtcdate.RTC_Month,rtcdate.RTC_Date,rtctime.RTC_Hours,rtctime.RTC_Minutes,rtctime.RTC_Seconds);

struct tm gmtime;
gmtime.tm_hour=rtctime.RTC_Hours;
gmtime.tm_min=rtctime.RTC_Minutes;
gmtime.tm_sec=rtctime.RTC_Seconds;
gmtime.tm_mday=rtcdate.RTC_Date;
gmtime.tm_mon=rtcdate.RTC_Month;
gmtime.tm_year=rtcdate.RTC_Year;
gmtime.tm_isdst=0;
tzname[0] = tzname[1] = "GMT";
setenv("TZ", "UTC", 1);
#ifdef _XOPEN_SOURCE
        timezone = 0;
        daylight = 0;
        #endif
tv->tv_sec = mktime(&gmtime);
tv->tv_usec = 0;
if(tz)
	{
	tz->tz_minuteswest = 0;
	tz->tz_dsttime = 0;
	}
return 0;
}


void rtc_settime(struct tm *gmtime)
{
RTC_TimeTypeDef rtctime;
RTC_DateTypeDef rtcdate;
rtctime.RTC_Hours=gmtime->tm_hour;
rtctime.RTC_Minutes=gmtime->tm_min;
rtctime.RTC_Seconds=gmtime->tm_sec;
rtcdate.RTC_Date=gmtime->tm_mday;
rtcdate.RTC_Month=gmtime->tm_mon;
rtcdate.RTC_Year=gmtime->tm_year;
rtcdate.RTC_WeekDay=gmtime->tm_wday;
rtc_set(&rtctime,&rtcdate);
}
