/*
 * *      saturation.c  ... firmware for inductor saturation meter based on  STM32F373VCT6 
 * *
 * *      Copyright (C) 2019 Jiri Pittner <jiri@pittnerovi.com>
 * *
 * *      This program is free software: you can redistribute it and/or modify
 * *      it under the terms of the GNU General Public License as published by
 * *      the Free Software Foundation, either version 3 of the License, or
 * *      (at your option) any later version.
 * *
 * *      This program is distributed in the hope that it will be useful,
 * *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 * *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * *      GNU General Public License for more details.
 * *      You should have received a copy of the GNU General Public License
 * *      along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * *
 * *
 * */

/* THEORY OF OPERATION
Inductor with internal resistance R and inductance  L at voltage source U:
U = RI + L dI/dt
Indicating explicitly functional dependences: U(t) = RI(t) + L(I) dI(t)/dt
for analytic solution assume U(t)=const, L(I)=const, then
I = U/R * (1-exp(-Rt/L))
In limit L=0 : I = U/R
In limit R=0 : I=U/L*t
However, as core saturates, L(I) is a general nonlinear function.
Moreover, U(t) is not constant since our capacitor bank is not infinite
We could measure U(t) along I(t), at the expense of halving the effective sample rate. (presently not implemented)
Note that R = Rsense + R_L if we take Vup as U
Since R_L should be small and can be measured by DC milliohmmeter, we take it as an optional
external correction (its determination from measurements by this device is difficult, since we do not have a constant
current source), with default 0.
We get L = (U-RI)/I' where I' is dI/dt
We sample I(t) at a equally spaced time grid with finite "dt" and approximate the derivative by finite difference
*/


//Makefile parameters:
//USART1=USART1_ON_PE01
//XTAL = 12000000ULL
//PLLMULT = 6


/*hardware:  cf. * PA11,12 = USB
 * PA13-15=JTAG
 *
 * PB0 = SW0
 * PB1 = SW1
 * PB2 =
 * PB3,4 = JTAG   
 * PB5=
 * PB6=
 * PB7=SDA
 * PB8=SCL
 * PB9=
 * PB10=
 * PB14=
 * PB15=
 *
 * PC0=  ADC10 Vsense0.01
 * PC1=  ADC11 Vsense0.1
 * PC2=  ADC12 Vsense0.01amplif
 * PC3=  ADC13 Vsense0.1amplif
 * PC4=  ADC14 Vup
 * PC5=  ADC15 Vtest
 * PC6=CS
 * PC7=SCK
 * PC8=MISO
 * PC9=MOSI
 * PC10=
 * PC11=
 * PC12=
 * PC13=
 * PC14=
 * PC15=
 *
 * PD0-7=LCD
 * PD8=
 * PD9=
 * PD10=
 * PD11=
 * PD12=
 * PD13=
 * PD14=
 * PD15=
 *
 * PE0=USART1TX
 * PE1=USART1RX
 * PE2=ROTL
 * PE3=ROTR
 * PE4=ROTP
 * PE5= 
 * PE6= 
 * PE7 = BUTL0
 * PE8 = BUTR1
 * PE9 = BUTU2
 * PE10= BUTD3
 * PE11 = BUTC4
 * PE12 =
 * PE13= 
 * PE14=
 * PE15=
 *
 * PF2=
 * PF3=
 * PF9=
 * PF10=
 *
 * */

//calibration constants
#define Rparasitic (0.02f+0.005f)
#define Rlmin 0.01f
#define Vdacoffset 0.18f
#define Voutmax 15.f
#define VCC3A 3.300f
#define R201 1200.f
#define R202 4700.f
#define R203 1200.f
#define R204 4700.f
#define Rsense0 0.01f
#define Rsense1 0.1f
#define R424 1200.f
#define R427 4700.f
#define R426 1200.f
#define R430 4700.f
#define R428 R426
#define R431 R430



#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <setjmp.h>
#include <errno.h>
#include <math.h>

#ifndef USART1_ON_PE01
#error USART1_ON_PE01 must be defined
#endif

#include "usart123.h"
#include "delay.h"
#include "lcd.h"
#include "stm32f37x_gpio.h"
#include "stm32f37x_rcc.h"
#include "stm32f37x_pwr.h"
#include "stm32f37x_tim.h"
#include "stm32f37x_exti.h"
#include "stm32f37x_dma.h"
#include "stm32f37x_adc.h"
#include "stm32f37x_dac.h"
#include "stm32f37x_misc.h"
#include "stm32f37x_syscfg.h"

#ifdef WITH_USB
#include "usb/inc/ttyacm.h"
#endif





volatile uint32_t milliseconds=0;
void TIM4_IRQHandler(void) __attribute__((optimize("-O1")));
void TIM4_IRQHandler(void)
{
TIM_ClearITPendingBit(TIM4,TIM_IT_Update);
++milliseconds;
}


void init_trigger(void)
{
GPIO_InitTypeDef GPIO_InitStructure;
RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOA, ENABLE);
GPIO_InitStructure.GPIO_Pin =  GPIO_Pin_0|GPIO_Pin_1;
GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
GPIO_Init(GPIOA, &GPIO_InitStructure);
GPIO_InitStructure.GPIO_Pin =  GPIO_Pin_2;
GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN;
GPIO_Init(GPIOA, &GPIO_InitStructure);
cbi(GPIOA,0);
sbi(GPIOA,1);
delay_ms(1);
cbi(GPIOA,1);
}




//leds - PA8-10
void init_led(void)
{
GPIO_InitTypeDef GPIO_InitStructure;
RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOA, ENABLE);
GPIO_InitStructure.GPIO_Pin =  GPIO_Pin_8|GPIO_Pin_9|GPIO_Pin_10;
GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
GPIO_Init(GPIOA, &GPIO_InitStructure);
}

void init_switches(void)
{
GPIO_InitTypeDef GPIO_InitStructure;
RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOB, ENABLE);
GPIO_InitStructure.GPIO_Pin =  GPIO_Pin_0|GPIO_Pin_1;
GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN;
GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
GPIO_Init(GPIOB, &GPIO_InitStructure);
}

void init_buttons(void)
{
GPIO_InitTypeDef GPIO_InitStructure;
RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOE, ENABLE);
GPIO_InitStructure.GPIO_Pin =  GPIO_Pin_7|GPIO_Pin_8|GPIO_Pin_11|GPIO_Pin_9|GPIO_Pin_10;
GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN;
GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
GPIO_Init(GPIOE, &GPIO_InitStructure);

//and interrupts
RCC_APB2PeriphClockCmd(RCC_APB2Periph_SYSCFG, ENABLE);
SYSCFG_EXTILineConfig(EXTI_PortSourceGPIOE,EXTI_PinSource7);
SYSCFG_EXTILineConfig(EXTI_PortSourceGPIOE,EXTI_PinSource8);
SYSCFG_EXTILineConfig(EXTI_PortSourceGPIOE,EXTI_PinSource9);
SYSCFG_EXTILineConfig(EXTI_PortSourceGPIOE,EXTI_PinSource10);
SYSCFG_EXTILineConfig(EXTI_PortSourceGPIOE,EXTI_PinSource11);

EXTI_InitTypeDef EXTI_InitStructure;
EXTI_InitStructure.EXTI_Line = EXTI_Line7|EXTI_Line8|EXTI_Line11|EXTI_Line9|EXTI_Line10;
EXTI_InitStructure.EXTI_Mode = EXTI_Mode_Interrupt;
EXTI_InitStructure.EXTI_Trigger = EXTI_Trigger_Rising_Falling;
EXTI_InitStructure.EXTI_LineCmd = ENABLE;
EXTI_Init(&EXTI_InitStructure);

NVIC_InitTypeDef NVIC_InitStructure;
NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 0x0d;
NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
NVIC_InitStructure.NVIC_IRQChannelSubPriority = 0x0d;
NVIC_InitStructure.NVIC_IRQChannel = EXTI15_10_IRQn;
NVIC_Init(&NVIC_InitStructure);
NVIC_InitStructure.NVIC_IRQChannelSubPriority = 0x0c;
NVIC_InitStructure.NVIC_IRQChannel = EXTI9_5_IRQn;
NVIC_Init(&NVIC_InitStructure);
}



void led(uint8_t which, uint8_t onoff) //which = 0-2 BGR
{
if(onoff) sbi(GPIOA,8+which); else cbi(GPIOA,8+which);
}

//working and display modes
typedef enum {ready=0,set_V,set_I,set_f,set_R,set_ns,set_nr,set_depth,measuring,results} WORKMODE;

//measurement ranges
#define  N_Im_range  16
static const float Im_scale[N_Im_range] = {0.1,0.15,0.22,0.33,0.47,0.68,1.,1.5,2.2,3.3,4.7,6.8,10.,15.,22.,33.};
#define N_freq_range  15
static const uint32_t freq_scale[N_freq_range] = {10,100,1000,10000,20000,40000,100000,200000,300000,400000,500000,600000,720000,800000,900000};

//measurement parameters
volatile WORKMODE workmode=ready;
volatile float Vt=8.f;
volatile int8_t Im_range=10;
volatile float Rl=Rlmin;
volatile int8_t freq_range=6;
volatile float Im;
volatile uint32_t freq;
volatile uint8_t sens=1;
volatile uint16_t nsampl=50;
volatile uint16_t nrep=32;
volatile uint8_t interrupt_measure=0;
volatile uint8_t average_depth=2;


void process_buttons(uint8_t which, uint32_t millisec)
{
//BUTL=0, BUTR=1, BUTC=4 (BUTD BUTU not installed)
//printf("button %d pressed for %ld\n",which,millisec);
switch (workmode)
	{
        case measuring: //interrupt measurement  by any button
		interrupt_measure=1;
		workmode=ready;
                break;
        case results:
		workmode=ready;
                break;
        case set_V:
		if(which==4) {workmode=set_I; break;}
		Vt += which==0? -1.f: 1.f;
		if(Vt<3.f) Vt=3.f;
		if(Vt>Voutmax) Vt=Voutmax;
                break;
        case set_I:
		if(which==4) {workmode=set_f; break;}
		if(which==0) --Im_range; else ++Im_range;
		if(Im_range<0) Im_range=0;
		if(Im_range>=N_Im_range) Im_range=N_Im_range-1;
		Im = Im_scale[Im_range];
                break;
        case set_f:
		if(which==4) {workmode=set_R; break;}
		if(which==0) --freq_range; else ++freq_range;
                if(freq_range<0) freq_range=0;
                if(freq_range>=N_freq_range) freq_range=N_freq_range-1;
                freq = freq_scale[freq_range];
                break;
        case set_R:
		if(which==4) {workmode=set_ns; break;}
		if(which==0) Rl /= sqrt(sqrt(2.)); else  Rl *= sqrt(sqrt(2.));
		if(Rl<Rlmin) Rl=Rlmin; //there is always capacitor ESR, mosfet, PCB, wire and connector resistances present besides the inductor's resistance
                break;
        case set_ns:
		if(which==4) {workmode=set_nr; break;}
		if(which==0) {if(nsampl>=100) nsampl -= 50;} else nsampl+= 50;
		if(nsampl>1000) nsampl=1000;
                break;
        case set_nr:
		if(which==4) {workmode=set_depth; break;}
		if(which==0) nrep >>= 1; else nrep <<= 1;
		if(nrep==0) nrep=1;
		if(nrep>1024) nrep=1024;
                break;
	case set_depth:
		if(which==4) {workmode=ready; break;}
		if(which==0) {if(average_depth>0) --average_depth;} else ++average_depth;
		if(average_depth>10) average_depth=10;
		break;
	case ready:
		if(which==4) workmode=measuring;
		else workmode=set_V;
		break;
	default:
		workmode=ready;
		break;
	}
}




void EXTI0_IRQHandler(void)
{
EXTI_ClearITPendingBit(EXTI_Line0);
}

void EXTI1_IRQHandler(void)
{
EXTI_ClearITPendingBit(EXTI_Line1);
}

void EXTI2_TS_IRQHandler(void)
{
EXTI_ClearITPendingBit(EXTI_Line2);
}

void EXTI3_IRQHandler(void) 
{
EXTI_ClearITPendingBit(EXTI_Line3);
}


void EXTI4_IRQHandler(void) //rotate push
{
EXTI_ClearITPendingBit(EXTI_Line4);
}

volatile uint32_t butpressed[5];

void EXTI9_5_IRQHandler(void) // button left,right, sw left
{
uint8_t which=0xff;
if(EXTI_GetITStatus(EXTI_Line7)==SET) which=0;
if(EXTI_GetITStatus(EXTI_Line8)==SET) which=1;
if(EXTI_GetITStatus(EXTI_Line9)==SET) which=2;
EXTI_ClearITPendingBit(EXTI_Line5);
EXTI_ClearITPendingBit(EXTI_Line6);
EXTI_ClearITPendingBit(EXTI_Line7);
EXTI_ClearITPendingBit(EXTI_Line8);
EXTI_ClearITPendingBit(EXTI_Line9);
if(which==0xff) return;
if(bit_is_set(GPIOE,7+which)) process_buttons(which,milliseconds-butpressed[which]);
else butpressed[which]=milliseconds;
}

void EXTI15_10_IRQHandler(void) //switch right
{
uint8_t which=0xff;
if(EXTI_GetITStatus(EXTI_Line10)==SET) which=3;
if(EXTI_GetITStatus(EXTI_Line11)==SET) which=4;
EXTI_ClearITPendingBit(EXTI_Line10);
EXTI_ClearITPendingBit(EXTI_Line11);
EXTI_ClearITPendingBit(EXTI_Line12);
EXTI_ClearITPendingBit(EXTI_Line13);
EXTI_ClearITPendingBit(EXTI_Line14);
EXTI_ClearITPendingBit(EXTI_Line15);
if(which==0xff) return;
if(bit_is_set(GPIOE,7+which)) process_buttons(which,milliseconds-butpressed[which]);
else butpressed[which]=milliseconds;
}

//////////////////////MEASUREMENT CONVERSIONS

uint16_t Vtest_to_dac1(float v)
{
float t=(Voutmax-v-Vdacoffset)*4096.f/VCC3A/(1.f+R202/R201);
return (uint16_t) t;
}

uint16_t I0thr_to_dac2(float i)
{
float t=i*4096./VCC3A*Rsense0;
return (uint16_t) t;
}

uint16_t I1thr_to_dac2(float i)
{
float t=i*4096./VCC3A*Rsense1;
return (uint16_t) t;
}

float Vtest_from_adc15(uint16_t a)
{
return VCC3A*a/4096.*(R203+R204)/R203;
}

float Vup_from_adc14(uint16_t a)
{
return VCC3A*a/4096.f*(R424+R427)/R424;
}

float I0_from_adc10(uint32_t a, uint16_t n)
{
return VCC3A*a/4096.f/Rsense0/n;
}

float I1_from_adc11(uint32_t a, uint16_t n)
{
return VCC3A*a/4096.f/Rsense1/n;
}

float I0_from_adc12(uint32_t a, uint16_t n)
{
return VCC3A*a/4096.f/Rsense0/(1.f+R430/R426)/n;
}

float I1_from_adc13(uint32_t a, uint16_t n)
{
return VCC3A*a/4096.f/Rsense1/(1.f+R431/R428)/n;
}




/////////////////////ADC ROUTINES

uint32_t read_ADC(uint16_t nrep)
{
uint32_t r=0;
//discard first after the switching
//ADC_SoftwareStartConv(ADC1);
while(ADC_GetFlagStatus(ADC1, ADC_FLAG_EOC) == RESET);
ADC_GetConversionValue(ADC1);

do{
	//ADC_SoftwareStartConv(ADC1);
	while(ADC_GetFlagStatus(ADC1, ADC_FLAG_EOC) == RESET);
	r+= ADC_GetConversionValue(ADC1);
}while(--nrep);
return r;
}


uint8_t init_ADC(uint8_t channel)
{
RCC_ADCCLKConfig(RCC_PCLK2_Div2);
RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOC, ENABLE);
GPIO_InitTypeDef    GPIO_InitStructure;
GPIO_InitStructure.GPIO_Pin = GPIO_Pin_0|GPIO_Pin_1|GPIO_Pin_2|GPIO_Pin_3|GPIO_Pin_4|GPIO_Pin_5;
GPIO_InitStructure.GPIO_Mode = GPIO_Mode_AN;
GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL ;
GPIO_Init(GPIOC, &GPIO_InitStructure);

RCC_APB2PeriphClockCmd(RCC_APB2Periph_ADC1, ENABLE);
ADC_DeInit(ADC1);

ADC_InitTypeDef     ADC_InitStructure;
ADC_StructInit(&ADC_InitStructure);
ADC_InitStructure.ADC_ScanConvMode = DISABLE;
ADC_InitStructure.ADC_ContinuousConvMode = ENABLE; //faster than always starting it
ADC_InitStructure.ADC_ExternalTrigConv = ADC_ExternalTrigConv_None;
ADC_InitStructure.ADC_DataAlign = ADC_DataAlign_Right;
ADC_InitStructure.ADC_NbrOfChannel = 1;
ADC_Init(ADC1, &ADC_InitStructure);

ADC_RegularChannelConfig(ADC1, ADC_Channel_Vrefint, 1, ADC_SampleTime_13Cycles5);
ADC_Cmd(ADC1, DISABLE);
ADC_TempSensorVrefintCmd(ENABLE);
delay_us(100);
ADC_Cmd(ADC1, ENABLE);

/* ADC1 calibration start */
ADC_ResetCalibration(ADC1);
while(ADC_GetResetCalibrationStatus(ADC1));
ADC_StartCalibration(ADC1);
while(ADC_GetCalibrationStatus(ADC1));
ADC_Cmd(ADC1, ENABLE);

ADC_RegularChannelConfig(ADC1, channel, 1, ADC_SampleTime_13Cycles5);
ADC_Cmd(ADC1, ENABLE);
return 0;
}

volatile uint8_t ADC_DMA_done;
volatile uint16_t ndma=0;
void DMA1_Channel1_IRQHandler(void)
{
if(DMA_GetITStatus(DMA1_IT_TC1) == SET )
	{
	DMA_ClearITPendingBit(DMA1_IT_TC1);
	//stop further conversions
	DMA_Cmd(DMA1_Channel1, DISABLE);
	TIM_Cmd(TIM19,DISABLE); 
	ADC_DMA_done=1;
	//printf("DMA done %d transfers\n",ndma);
	}
}
 
void TIM19_IRQHandler(void) __attribute__((optimize("-O1")));
void TIM19_IRQHandler(void)
{
if(TIM_GetITStatus(TIM19,TIM_IT_Update) == SET )
	{
	TIM_ClearITPendingBit(TIM19,TIM_IT_Update);
	++ndma;
	}
}


//Tconv = Sampling time + 12.5 cycles of PCLK2_Div2 (max 14MHz)
//fastest is  ADC_SampleTime_1Cycles5 -> 14 cycles of XTAL*PLLMULT/divisor
//1.17us fastest
//maximum samling frequency is thus 850khz
uint8_t init_ADC_DMA(uint8_t channel, uint8_t ADC_SampleTime, uint32_t sampling_freq, uint16_t *buffer, uint16_t blocksize)
{
ndma=0;
RCC_ADCCLKConfig(RCC_PCLK2_Div6); //get below 14MHz or try overclocking by Div4 to 18MHz - works at cost of accuracy???
RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOC, ENABLE);
GPIO_InitTypeDef    GPIO_InitStructure;
GPIO_InitStructure.GPIO_Pin = GPIO_Pin_0|GPIO_Pin_1|GPIO_Pin_2|GPIO_Pin_3|GPIO_Pin_4|GPIO_Pin_5;
GPIO_InitStructure.GPIO_Mode = GPIO_Mode_AN;
GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL ;
GPIO_Init(GPIOC, &GPIO_InitStructure);

RCC_APB2PeriphClockCmd(RCC_APB2Periph_ADC1, ENABLE);
ADC_DeInit(ADC1);

ADC_InitTypeDef     ADC_InitStructure;
ADC_StructInit(&ADC_InitStructure);
ADC_InitStructure.ADC_ScanConvMode = DISABLE;
ADC_InitStructure.ADC_ContinuousConvMode = DISABLE;
ADC_InitStructure.ADC_ExternalTrigConv = ADC_ExternalTrigConv_T19_TRGO;
ADC_InitStructure.ADC_DataAlign = ADC_DataAlign_Right;
ADC_InitStructure.ADC_NbrOfChannel = 1;
ADC_Init(ADC1, &ADC_InitStructure);


ADC_RegularChannelConfig(ADC1, ADC_Channel_Vrefint, 1, ADC_SampleTime);
ADC_Cmd(ADC1, DISABLE);
ADC_TempSensorVrefintCmd(ENABLE);
delay_us(100);
ADC_Cmd(ADC1, ENABLE);

/* ADC1 calibration start */
ADC_ResetCalibration(ADC1);
while(ADC_GetResetCalibrationStatus(ADC1));
ADC_StartCalibration(ADC1);
while(ADC_GetCalibrationStatus(ADC1));
ADC_Cmd(ADC1, ENABLE);

ADC_RegularChannelConfig(ADC1, channel, 1, ADC_SampleTime);

ADC_ExternalTrigConvCmd(ADC1,ENABLE);

//DMA
RCC_AHBPeriphClockCmd(RCC_AHBPeriph_DMA1, ENABLE);
static DMA_InitTypeDef DMA_InitStruct;

  DMA_StructInit(&DMA_InitStruct);
  DMA_InitStruct.DMA_PeripheralBaseAddr = (uint32_t)&(ADC1->DR);
  DMA_InitStruct.DMA_MemoryBaseAddr = (uint32_t)buffer;
  DMA_InitStruct.DMA_DIR = DMA_DIR_PeripheralSRC;
  DMA_InitStruct.DMA_BufferSize = blocksize;
  DMA_InitStruct.DMA_PeripheralInc = DMA_PeripheralInc_Disable;
  DMA_InitStruct.DMA_MemoryInc = DMA_MemoryInc_Enable;
  DMA_InitStruct.DMA_PeripheralDataSize = DMA_PeripheralDataSize_HalfWord;
  DMA_InitStruct.DMA_MemoryDataSize = DMA_MemoryDataSize_HalfWord;
  DMA_InitStruct.DMA_Mode = DMA_Mode_Normal;
  DMA_InitStruct.DMA_Priority = DMA_Priority_Medium;
  DMA_InitStruct.DMA_M2M = DMA_M2M_Disable;
  DMA_Init(DMA1_Channel1, &DMA_InitStruct);


DMA_Cmd(DMA1_Channel1, DISABLE);
DMA_ClearFlag(DMA1_FLAG_TC1); //clear transfer complete to allow next one
DMA_ClearFlag(DMA1_FLAG_TE1); //clear transfer error
DMA_ITConfig(DMA1_Channel1,DMA_IT_TC,ENABLE);
DMA_Cmd(DMA1_Channel1, ENABLE);

//timer19 for sampling control
RCC_APB2PeriphClockCmd(RCC_APB2Periph_TIM19, ENABLE);
RCC_APB2PeriphResetCmd(RCC_APB2Periph_TIM19, ENABLE);
RCC_APB2PeriphResetCmd(RCC_APB2Periph_TIM19, DISABLE);
TIM_InternalClockConfig(TIM19);
TIM_TimeBaseInitTypeDef TIM_TimeBaseInitStructure;
TIM_TimeBaseStructInit(&TIM_TimeBaseInitStructure);
TIM_TimeBaseInitStructure.TIM_Prescaler=0; //(divided by 1+prescaler)
TIM_TimeBaseInitStructure.TIM_Period = (XTAL*PLLMULT)/(0+1)/sampling_freq;
TIM_TimeBaseInitStructure.TIM_CounterMode = TIM_CounterMode_Up;
TIM_TimeBaseInit(TIM19,&TIM_TimeBaseInitStructure);
TIM19->CNT=0;
TIM_SelectOutputTrigger(TIM19,TIM_TRGOSource_Update);
TIM_ITConfig(TIM19, TIM_IT_Update, ENABLE);
//TIM_Cmd(TIM19,ENABLE);



NVIC_InitTypeDef NVIC_InitStructure;

NVIC_InitStructure.NVIC_IRQChannel = TIM19_IRQn;
NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 0;
NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
NVIC_Init(&NVIC_InitStructure);

NVIC_InitStructure.NVIC_IRQChannel = DMA1_Channel1_IRQn;
NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 0;
NVIC_InitStructure.NVIC_IRQChannelSubPriority = 0;
NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
NVIC_Init(&NVIC_InitStructure);


return 0;
}

void start_ADC_DMA()
{
ADC_DMA_done=0;
ADC_DMACmd(ADC1,ENABLE);
//ADC_SoftwareStartConv(ADC1);
ADC_Cmd(ADC1, ENABLE);
TIM19->CNT=0;
TIM_Cmd(TIM19,ENABLE);
}


//////////////////////DAC ROUTINES
uint8_t init_set_DAC(uint16_t data1,uint16_t data2)
{
RCC_APB1PeriphClockCmd(RCC_APB1Periph_DAC1, ENABLE);
RCC_APB1PeriphClockCmd(RCC_APB1Periph_DAC2, ENABLE);
RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOA, ENABLE);
GPIO_InitTypeDef    GPIO_InitStructure;
GPIO_InitStructure.GPIO_Pin = GPIO_Pin_4|GPIO_Pin_6;
GPIO_InitStructure.GPIO_Mode = GPIO_Mode_AN;
GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL ;
GPIO_Init(GPIOA, &GPIO_InitStructure);
RCC_APB1PeriphClockCmd(RCC_APB1Periph_DAC1, ENABLE);
RCC_APB1PeriphClockCmd(RCC_APB1Periph_DAC2, ENABLE);
DAC_DeInit(DAC1);
DAC_DeInit(DAC2);

DAC_InitTypeDef DAC_InitStruct;
DAC_InitStruct.DAC_Trigger = DAC_Trigger_None;
DAC_InitStruct.DAC_WaveGeneration = DAC_WaveGeneration_None;
DAC_InitStruct.DAC_LFSRUnmask_TriangleAmplitude = DAC_LFSRUnmask_Bit0;
DAC_InitStruct.DAC_OutputBuffer = DAC_OutputBuffer_Enable;
DAC_Init(DAC1,1,&DAC_InitStruct);
DAC_Init(DAC2,1,&DAC_InitStruct);
DAC_Cmd(DAC1,DAC_Channel_1,ENABLE);
DAC_Cmd(DAC2,DAC_Channel_1,ENABLE);
DAC_SetChannel1Data(DAC1,DAC_Align_12b_R,data1); //this sets the data
DAC_SetChannel1Data(DAC2,DAC_Align_12b_R,data2); //this sets the data
//printf("set_DAC %u %u\n",data1,data2);
return 0;
}

//for milliseconds
void init_tim4(void)
{
RCC_APB1PeriphClockCmd(RCC_APB1Periph_TIM4, ENABLE);
RCC_APB1PeriphResetCmd(RCC_APB1Periph_TIM4, ENABLE);
RCC_APB1PeriphResetCmd(RCC_APB1Periph_TIM4, DISABLE);
TIM_TimeBaseInitTypeDef TIM_TimeBaseInitStructure;
TIM_TimeBaseStructInit(&TIM_TimeBaseInitStructure);
TIM_TimeBaseInitStructure.TIM_Prescaler=1;
TIM_TimeBaseInitStructure.TIM_Period = (XTAL*PLLMULT)/2/1000;
TIM_TimeBaseInitStructure.TIM_CounterMode = TIM_CounterMode_Up;
TIM_TimeBaseInit(TIM4,&TIM_TimeBaseInitStructure);
TIM_InternalClockConfig(TIM4);
TIM4->CNT=0;
TIM_ITConfig(TIM4, TIM_IT_Update, ENABLE);
TIM_Cmd(TIM4,ENABLE);
NVIC_InitTypeDef NVIC_InitStructure;
NVIC_InitStructure.NVIC_IRQChannel = TIM4_IRQn;
NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 0;
NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
NVIC_Init(&NVIC_InitStructure);
}


//[amplif][sense]
static const uint8_t I_channel[2][2] = {{ADC_Channel_10,ADC_Channel_11},{ADC_Channel_12,ADC_Channel_13}};
static const float (*I_conversion[2][2])(uint32_t,uint16_t) = {{&I0_from_adc10,&I1_from_adc11},{&I0_from_adc12,&I1_from_adc13}};

uint8_t do1measurement(uint8_t sense, uint8_t amplification, uint32_t sampling_freq, float Vtest, uint32_t *abuffer, uint16_t blocksize, float *Vtest_eff, uint16_t *maxgoodsamples)
{
uint16_t bbuffer[blocksize];
uint8_t r=0;
init_trigger();
//check Vtest whether charged
init_ADC(ADC_Channel_15);
float Vtest_readback = Vtest_from_adc15(read_ADC(10))/10.;
while(fabs(Vtest_readback-Vtest)>0.5) {delay_ms(200);led(1,1); delay_ms(200);led(1,0);}
printf("Vtest set %5.2f readback %5.2f ",Vtest,Vtest_readback);
	

init_ADC_DMA(I_channel[amplification][sense],ADC_SampleTime_1Cycles5,sampling_freq,bbuffer,blocksize);
//printf("TRIGGER ON!\n");
start_ADC_DMA();
sbi(GPIOA,0); //trigger current
while(ndma<blocksize || !ADC_DMA_done) //the latter should be actually enough
	{
	if(bit_is_clear(GPIOA,2)) //I_OK false overcurrent condition 
		{
		cbi(GPIOA,0);
		*maxgoodsamples=ndma;
		//printf("OVERCURRENT TRIGGER OFF!\n");
		led(2,1); //blue
		r=1;
		delay_us(5);
		sbi(GPIOA,1);
		delay_us(5);
		cbi(GPIOA,1);
		}
	}
cbi(GPIOA,0);//trigger off
//if(r==0) printf("NORMAL TRIGGER OFF!\n");
init_ADC(ADC_Channel_15);
float Vtest_left = Vtest_from_adc15(read_ADC(10))/10.;
printf("after discharge %5.2f\n",Vtest_left);
*Vtest_eff = 0.5*(Vtest_left+Vtest_readback);
if(Vtest_readback-Vtest_left>0.8) led(1,1); //green indicate big voltage drop
uint16_t i;
for(i=0; i<blocksize; ++i) abuffer[i] += bbuffer[i];
return r;
}


//convert ADC measurement to floating point current
uint8_t adc2i(uint8_t sense, uint8_t amplification, uint32_t *abuffer,float *Ibuffer, uint16_t blocksize, uint16_t *validsize,uint16_t nrepeat)
{
uint16_t i;
*validsize = blocksize;
for(i=0; i<blocksize; ++i) 
	{
	Ibuffer[i] = (*I_conversion[amplification][sense])(abuffer[i],nrepeat);
	if(i>10 && i<blocksize-2&& Ibuffer[i] < Ibuffer[i-1] && Ibuffer[i+1] < Ibuffer[i-1] &&  Ibuffer[i+2] < Ibuffer[i-1]) {*validsize=i-1; break;} //current must be a monotonic growing function of time
	}
return 0;
}



void analyze_results(uint32_t sampling_freq, float *Ibuffer, float *Ubuffer, float Uconst, float R, uint16_t nsamples, float *Lbuffer)
{
float dt=1./sampling_freq;
uint16_t j;
for(j=1; j<nsamples; ++j)
	{
	float I=.5*(Ibuffer[j]+Ibuffer[j-1]);
	float Ider = (Ibuffer[j]-Ibuffer[j-1])/dt;
	float U;
	if(!Ubuffer) U=Uconst;
	else U=.5*(Ubuffer[j]+Ubuffer[j-1]);
	float L = (U - R*I)/Ider; //account for the voltage drop on the shunt and parasitic resistors
	Lbuffer[j-1]=L;
	Ibuffer[j-1]=I;
	if(Ubuffer) Ubuffer[j-1]=U;
	}
}


uint8_t performmeasurement(uint8_t sense, uint8_t amplification, uint32_t sampling_freq, float Vtest, float Imax, uint16_t nsamples, float RL, uint16_t nrepeat, float *Lx0, float *Ix0, float *Lx1, float *Ix1, float *Lx2, float *Ix2, float *Lx3, float *Ix3)
{
printf("Commence measurement: sensor %d amplification %d samplig freq %lu Vtest %f Imax %f nsamples %d RL %f repetitions %d\n",sense,amplification,sampling_freq,Vtest,Imax,nsamples,RL,nrepeat);

interrupt_measure=0;
float Ibuffer[nsamples];
float Lbuffer[nsamples];
uint16_t validsize;
float Vtest_effective;
float Vtest_eff;
uint16_t i;
{
uint32_t abuffer[nsamples];

uint16_t dac1,dac2;
dac1= Vtest_to_dac1(Vtest);
if(dac1>4095) {printf("Voltage DAC1 out of range: %d\n",dac1); dac1=4095;} 
dac2= sense ? I1thr_to_dac2(Imax) : I0thr_to_dac2(Imax);
if(dac2>4095) {printf("Imax DAC2 out of range: %d\n",dac2); dac2=4095;} 
init_set_DAC(dac1,dac2);
delay_ms(300);//charge the large capacitor bank to Vtest 

//zero the accumulation buffer
memset(abuffer,0,nsamples*sizeof(uint32_t));
Vtest_effective=0.;
uint16_t maxgood=nsamples;
for(i=0; i<nrepeat; ++i)
	{
	lcd_cursor(1,0);
	char line[24];
        sprintf(line,"%4d of %4d    ",i,nrepeat);
	line[16]=0;
        lcd_print(line);
	uint16_t maxgood1=nsamples;
	do1measurement(sense,amplification,sampling_freq,Vtest,abuffer,nsamples,&Vtest_eff, &maxgood1);
	if(maxgood1 < maxgood) maxgood=maxgood1;
	if(interrupt_measure) return 2;
	Vtest_effective += Vtest_eff;
	delay_ms(200); //recharge cap bank
	}
Vtest_effective /= nrepeat;

adc2i(sense,amplification,abuffer,Ibuffer,nsamples,&validsize,nrepeat);
if(maxgood<validsize) validsize=maxgood;
}

if(validsize<2) return 3;
//try smoothing the data before L is computed???

analyze_results(sampling_freq,Ibuffer,NULL,Vtest_effective,Rparasitic+RL+(sense==0?Rsense0:Rsense1),validsize,Lbuffer);
--validsize;
if(validsize<2) return 3;
float dLdIbuffer[validsize];

for(i=1; i<validsize; ++i) dLdIbuffer[i-1] = (Lbuffer[i]-Lbuffer[i-1])/(Ibuffer[i]-Ibuffer[i-1]);
dLdIbuffer[validsize-1] = dLdIbuffer[validsize-2];


for(i=0; i<validsize; ++i) if(Ibuffer[i]>Imax) {validsize=i-1; break;}

uint16_t naverage=validsize;
float Iaverage[naverage];
float Laverage[naverage];
memcpy(Iaverage,Ibuffer,naverage*sizeof(float));
memcpy(Laverage,Lbuffer,naverage*sizeof(float));
uint8_t k;
for(k=0; k<average_depth; ++k)
	{
	for(i=1; i<naverage; ++i)
		{
		float t= (Iaverage[i-1]+Iaverage[i])*0.5f;
		Iaverage[i-1] = t;
		t= (Laverage[i-1]+Laverage[i])*0.5f;
		Laverage[i-1] = t;
		}
	if(naverage==0) return 3;
	--naverage;
	}


//print all curve
printf("Raw data\n");
for(i=0; i<validsize; ++i)
	{
	printf("I= %06.3f L= %8.6f dL/dI= %g\n",Ibuffer[i],Lbuffer[i],dLdIbuffer[i]);
	}

printf("Smoothened data\n");
for(i=0; i<naverage; ++i)
        {
        printf("I= %06.3f L= %8.6f\n",Iaverage[i],Laverage[i]);
        }



//if wrong results set to NaN and return 1
uint8_t error=0;
for(i=0; i<naverage; ++i) if(Laverage[i]<0.f) error=1;
if(error) return 1;
	
//select most important results to display in two lines

//Maximum inductance
*Lx0=0;
*Ix0=0;
for(i=0; i<naverage; ++i)
	{
	if(Laverage[i]> *Lx0) {*Lx0=Laverage[i]; *Ix0= Iaverage[i];}
	}

//At which max current at least half of that inductance is remaining
float t= *Lx0 * .5f;
*Lx1=0.f;
for(i=naverage-1; i>0; --i)
        {
        if(Laverage[i]>=t) {*Lx1=Laverage[i]; *Ix1= Iaverage[i]; break;}
        }
if(*Lx1==0.f) {*Lx1=Laverage[naverage-1]; *Ix1= Iaverage[naverage-1];}

//Minimum inductance
*Lx2=1e6;
*Ix2=0;
for(i=0; i<naverage; ++i)
        {
        if(Laverage[i]< *Lx2 && Laverage[i]>=0 && Iaverage[i]<=Imax) {*Lx2=Laverage[i]; *Ix2= Iaverage[i];}
        }

//Average inductance at currents <=Ix1
*Lx3=0.f;
*Ix3= *Ix1;
uint16_t nx=0;
for(i=0; i<naverage; ++i)
        {
        if(Iaverage[i] <= *Ix3) {*Lx3+=Laverage[i]; ++nx;}
        }
*Lx3 /= nx;

return 0;
}



///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int main(void)
{
FILE *us1=fopen("usart 1 115200 1","w+"); setlinebuf(us1);
stdout=stderr=us1;
printf("Reset!\n");
lcd_init(); 
lcd_light(0);
lcd_print("Reset");
init_tim4();
init_led();
led(0,0);led(1,0);led(2,0);
init_buttons();
init_switches();
delay_ms(20); workmode=ready; //avoid spurious button signals after their init
printf("Init done!\n");

#ifdef WITH_USB
FILE *usb=fopen("usb 1","w+"); 
if(usb)
	{
	setlinebuf(usb);
	printf("USB Init success!\n");
	//does not mean it is connected to host yet!
	}
else	printf("USB Init failed!\n");
#endif

uint8_t amplif;
freq=freq_scale[freq_range];
Im=Im_scale[Im_range];

while(1)
{
char line[24];
float Lx0,Ix0,Lx1,Ix1,Lx2,Ix2,Lx3,Ix3;

if(workmode==ready) {led(0,0);led(1,0);led(2,0);}
if(workmode!=results) lcd_light(0); else lcd_light(1);

//read switches and adjust Imax accordingly
if(bit_is_set(GPIOB,0)) sens=0; else sens=1;
float Imeff = Im * (sens==0? Rsense1/Rsense0 : 1.f);

switch(workmode)
	{
	default:
	case ready:
		lcd_cursor(0,0);
		sprintf(line,"%04.1fV %04.1fA %03luK",Vt,Imeff,freq/1000);
		lcd_print(line);
		lcd_cursor(1,0);
		lcd_print("Press C to start");
		break;
	case measuring:
		{
		uint8_t r;
		//amplification choose based on Im and sens
		amplif=0;
		float vdropmax = Im * Rsense1; 
		if(vdropmax * (1.f+R431/R428) < VCC3A) amplif=1; //will be within range after amplification
		r=performmeasurement(sens,amplif,freq,Vt,Imeff,nsampl,Rl,nrep,&Lx0,&Ix0,&Lx1,&Ix1,&Lx2,&Ix2,&Lx3,&Ix3);
		workmode = r==2?ready:results;
		if(r) //error
			{
			led(0,1); //red indicate error condition
			Lx0=Ix0=Lx1=Ix1=Lx2=Ix2=Lx3=Ix3= (0.0f / 0.0f);
			}
		}
		break;
	case results:
		lcd_cursor(0,0);
		if((milliseconds & 4096)==4096) sprintf(line,"%.3fmH @ %.2fA  ",1000.f*Lx0,Ix0);
		else sprintf(line,"%.3fmH @ %.2fA  ",1000.f*Lx2,Ix2);
		line[16]=0;
                lcd_print(line);
                lcd_cursor(1,0);
		if((milliseconds & 4096)==4096) sprintf(line,"%.3fmH @ %.2fA  ",1000.f*Lx1,Ix1);
		else sprintf(line,"%.3fmH @<%.2fA  ",1000.f*Lx3,Ix3);
		line[16]=0;
                lcd_print(line);
		break;
	case set_V:
		lcd_cursor(0,0); lcd_print("Test Voltage    ");
		lcd_cursor(1,0);
		sprintf(line,"V=%fV        ",Vt); line[16]=0;
                lcd_print(line);
		break;
	case set_I: 
		lcd_cursor(0,0); lcd_print("Current Limit   ");
                lcd_cursor(1,0);
                sprintf(line,"I=%.1fA        ",Imeff); line[16]=0;
                lcd_print(line);
		break;
	case set_f:
		lcd_cursor(0,0); lcd_print("Sampling freq.  ");
                lcd_cursor(1,0);
                sprintf(line,"f=%luHz        ",freq); line[16]=0;
                lcd_print(line);
		break;
	case set_R:
		lcd_cursor(0,0); lcd_print("Parasitic R     ");
                lcd_cursor(1,0);
                sprintf(line,"R=%.3fOhm     ",Rl); line[16]=0;
                lcd_print(line);
                break;
	case set_ns:
		lcd_cursor(0,0); lcd_print("No. of samples  ");
                lcd_cursor(1,0);
                sprintf(line,"N=%u        ",nsampl); line[16]=0;
                lcd_print(line);
                break;
	case set_nr:
		lcd_cursor(0,0); lcd_print("No. of repetitions");
                lcd_cursor(1,0);
                sprintf(line,"n=%u        ",nrep); line[16]=0;
                lcd_print(line);
		break;
	case set_depth:
                lcd_cursor(0,0); lcd_print("Averaging depth  ");
                lcd_cursor(1,0);
                sprintf(line,"m=%u        ",average_depth); line[16]=0;
                lcd_print(line);
		break;
	}
//check input from usart1 
if(inuart1line
#ifdef WITH_USB
	||inusbline
#endif
	) //select() not implemented yet
	{
	char line[256];
	char *p=NULL;
	if(inuart1line) {p=fgets(line,256,us1); stdout=us1;}
#ifdef WITH_USB
	if(inusbline) {p=fgets(line,256,usb); stdout=usb;}
#endif
	if(p) 
	 {
	 int n=strlen(p);
	 if(p[n-1]=='\n') p[n-1]=0;
	 switch(line[0]) {
		case 'v':
			sscanf(line+1,"%f",&Vt);
			break;
		case 'i':
			sscanf(line+1,"%f",&Im);
                        break;
		case 'R':
			sscanf(line+1,"%f",&Rl);
                        break;
		case 'f':
			scanf(line+1,"%ld",&freq);
			break;
		case 'N':
			{
			uint32_t n;
			scanf(line+1,"%ld",&n);
			nsampl=n;
			}
			break;
		case 'n':
			{
                        uint32_t n;
                        scanf(line+1,"%ld",&n);
                        nrep=n;
                        }
                        break;
		case 'a':
			{
                        uint32_t n;
                        scanf(line+1,"%ld",&n);
                        average_depth=n;
                        }
                        break;
		case 'r':
			workmode=measuring;
			break;
#ifdef WITH_USB
		case 'u':
			fprintf(usb,line);
			break;
#endif
		default:
			printf("Unknown command %s\n",line);
			break;	
	}}
	}

delay_ms(100);
}//while
}//main

