/*
 * *      ttyacm.c  ... USB virtual serial port for STM32
 * *
 * *      Copyright (C) 2019 Jiri Pittner <jiri@pittnerovi.com>
 * *
 * *      This program is free software: you can redistribute it and/or modify
 * *      it under the terms of the GNU General Public License as published by
 * *      the Free Software Foundation, either version 3 of the License, or
 * *      (at your option) any later version.
 * *
 * *      This program is distributed in the hope that it will be useful,
 * *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 * *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * *      GNU General Public License for more details.
 * *      You should have received a copy of the GNU General Public License
 * *      along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * *
 * *
 * *      It employs STM32 HAL and USB library, which are under a different STM license
 *        and NOT included as a part of this program
 * */



#include "ttyacm.h"
#include "stm32f37x.h"
#include "usbd_core.h"
#include "usbd_cdc.h"
#include "usbd_cdc_desc.h"
#include "peripherals/stm32f37x_gpio.h"
#include "peripherals/stm32f37x_rcc.h"

extern USBD_DescriptorsTypeDef VCP_Desc;
extern USBD_CDC_ItfTypeDef  USBD_CDC_fops;

//receive buffer
volatile uint8_t ttyacmRxbuffer[USB_FS_MAX_PACKET_SIZE];
volatile uint8_t inusb[INUSBMAX+1];
volatile uint16_t inusblen=0;
volatile uint16_t inusbline=0;
volatile uint8_t ttyacm_initialized=0;
volatile uint8_t ttyacm_connected=0;
volatile uint8_t usbecho=1;

USBD_HandleTypeDef hUSBDDevice;

static int8_t ttyacm_Init(void)
{
//printf("USB connected!\n");
ttyacm_connected=1;
return USBD_CDC_SetRxBuffer(&hUSBDDevice, ttyacmRxbuffer);
}

static int8_t ttyacm_DeInit(void)
{
ttyacm_connected=0;
return (USBD_OK);
}

static int8_t ttyacm_Control(uint8_t cmd, uint8_t* pbuf, uint16_t length)
{
	inusbline=inusblen=0;

	//printf("ttyacm_Control cmd=%d\n",cmd);
	if(cmd==CDC_GET_LINE_CODING)//fake something
		{
		uint32_t t=115200;
		memcpy(pbuf,&t,sizeof(uint32_t));
		pbuf[4] = 0;
		pbuf[5] = 0;
		pbuf[6] = 0;
		}
	return (USBD_OK);
	}

static int8_t ttyacm_Receive(uint8_t* pbuf, uint32_t *Len)
{

	uint32_t len= *Len;
	if(inusblen+len>= INUSBMAX) len=INUSBMAX-inusblen;
	char *p0 = pbuf;
	char *p = inusb+inusblen;
	for(int i=0; i<len; ++i)
		{
		int newline=0;
		if(*p0=='\r' || *p0=='\n') {newline=1; *p ='\n';}
		else *p = *p0;
		ttyacm_Transmit(p,1); //echo the input back to ttyacm
		++p; ++p0;
		if(inusbline) continue; //ignore further input
		inusblen++;
		if(newline || inusblen == INUSBMAX)
			{
			*p = 0; inusbline=inusblen;
			}
		}


//ask for next packet
	USBD_CDC_ReceivePacket(&hUSBDDevice);
	return USBD_OK;
}


USBD_CDC_ItfTypeDef USBD_CDC_fops =
	{
	  ttyacm_Init,
	  ttyacm_DeInit,
	  ttyacm_Control,
	  ttyacm_Receive
	};

uint8_t ttyacm_Transmit(uint8_t *buf, uint16_t len)
{
	//copy to a local buffer is not needed here, USBD copies data to its own endpoint buffer for DMA transfer
	if(!ttyacm_connected) return 1;
	if(USBD_CDC_SetTxBuffer(&hUSBDDevice, buf, len) != USBD_OK) return 2;
	if(USBD_CDC_TransmitPacket(&hUSBDDevice) != USBD_OK) return 3;
	return 0;
}

	//main initialization routine
uint8_t ttyacm_init(void)
{
	ttyacm_initialized=0;
	  /* Configure USB DM and DP pins */
	  GPIO_InitTypeDef  GPIO_InitStruct;
	  RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOA, ENABLE);
	  GPIO_InitStruct.GPIO_Pin = (GPIO_Pin_11 | GPIO_Pin_12);
	  GPIO_InitStruct.GPIO_Mode = GPIO_Mode_AF;
	  GPIO_InitStruct.GPIO_PuPd = GPIO_PuPd_NOPULL;
	  GPIO_InitStruct.GPIO_Speed = GPIO_Speed_50MHz;
	  GPIO_Init(GPIOA, &GPIO_InitStruct);
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource11, GPIO_AF_14);
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource12, GPIO_AF_14);


	if(USBD_OK != USBD_Init(&hUSBDDevice, &VCP_Desc, 0)) return 1;
	if(USBD_OK != USBD_RegisterClass(&hUSBDDevice, &USBD_CDC)) return 2;
	if(USBD_OK != USBD_CDC_RegisterInterface(&hUSBDDevice, &USBD_CDC_fops)) return 3;
	if(USBD_OK != USBD_Start(&hUSBDDevice)) return 5;

	ttyacm_initialized=1;
	ttyacm_connected=0;
inusbline=inusblen=0;
return 0;
}


void my_usb_final_init(void)
{
//assumed 72MHz system
RCC_USBCLKConfig(RCC_USBCLKSource_PLLCLK_1Div5);
}
